#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import traceback
import uuid
from typing import Iterable
from urllib.parse import quote

import requests
from pydantic import SecretStr

from metadata.generated.schema.api.lineage.addLineage import AddLineage
from metadata.generated.schema.entity.data.dashboard import Dashboard as Model_Dashboard
from metadata.generated.schema.entity.data.table import Table
from metadata.generated.schema.entity.services.dashboardService import (
    DashboardServiceType,
)
from metadata.generated.schema.type.entityLineage import EntitiesEdge
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import (
    ConfigModel,
    Entity,
    IncludeFilterPattern,
    WorkflowContext,
)
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.table_metadata import Chart, Dashboard
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_alchemy_helper import SQLSourceStatus
from metadata.ingestion.source.sql_source import SQLSourceStatus
from metadata.utils.helpers import get_dashboard_service_or_create

HEADERS = {"Content-Type": "application/json", "Accept": "*/*"}


logger: logging.Logger = logging.getLogger(__name__)


class MetabaseSourceConfig(ConfigModel):
    username: str
    password: SecretStr
    host_port: str
    dashboard_pattern: IncludeFilterPattern = IncludeFilterPattern.allow_all()
    chart_pattern: IncludeFilterPattern = IncludeFilterPattern.allow_all()
    service_name: str
    service_type: str = "Metabase"
    database_service_name: str = None

    def get_connection_url(self):
        pass


class MetabaseSource(Source[Entity]):
    config: MetabaseSourceConfig
    metadata_config: MetadataServerConfig
    status: SQLSourceStatus

    def __init__(
        self,
        config: MetabaseSourceConfig,
        metadata_config: MetadataServerConfig,
        ctx: WorkflowContext,
    ):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.status = SQLSourceStatus()
        params = {}
        params["username"] = self.config.username
        params["password"] = self.config.password.get_secret_value()
        try:
            resp = requests.post(
                self.config.host_port + "/api/session/",
                data=json.dumps(params),
                headers=HEADERS,
            )
        except Exception as err:
            raise ConnectionError(f"{err}")
        session_id = resp.json()["id"]
        self.metabase_session = {"X-Metabase-Session": session_id}
        self.dashboard_service = get_dashboard_service_or_create(
            config.service_name,
            DashboardServiceType.Metabase.name,
            config.username,
            config.password.get_secret_value(),
            config.host_port,
            metadata_config,
        )
        self.charts = []
        self.metric_charts = []

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = MetabaseSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def next_record(self) -> Iterable[Entity]:
        yield from self.get_dashboards()

    def get_charts(self, charts) -> Iterable[Chart]:
        for chart in charts:
            try:
                chart_details = chart["card"]
                if not self.config.chart_pattern.included(chart_details["name"]):
                    self.status.filter(chart_details["name"])
                    continue
                yield Chart(
                    id=uuid.uuid4(),
                    name=chart_details["name"],
                    displayName=chart_details["name"],
                    description=chart_details["description"]
                    if chart_details["description"] is not None
                    else "",
                    chart_type=str(chart_details["display"]),
                    url=self.config.host_port,
                    service=EntityReference(
                        id=self.dashboard_service.id, type="dashboardService"
                    ),
                )
                self.charts.append(chart_details["name"])
                self.status.scanned(chart_details["name"])
            except Exception as err:
                logger.error(repr(err))
                traceback.print_exc()
                continue

    def get_dashboards(self):
        resp_dashboards = self.req_get("/api/dashboard")
        if resp_dashboards.status_code == 200:
            for dashboard in resp_dashboards.json():
                resp_dashboard = self.req_get(f"/api/dashboard/{dashboard['id']}")
                dashboard_details = resp_dashboard.json()
                self.charts = []
                if not self.config.dashboard_pattern.included(
                    dashboard_details["name"]
                ):
                    self.status.filter(dashboard_details["name"])
                    continue
                yield from self.get_charts(dashboard_details["ordered_cards"])
                yield Dashboard(
                    id=uuid.uuid4(),
                    name=dashboard_details["name"],
                    url=self.config.host_port,
                    displayName=dashboard_details["name"],
                    description=dashboard_details["description"]
                    if dashboard_details["description"] is not None
                    else "",
                    charts=self.charts,
                    service=EntityReference(
                        id=self.dashboard_service.id, type="dashboardService"
                    ),
                )
                yield from self.get_lineage(
                    dashboard_details["ordered_cards"], dashboard_details["name"]
                )

    def get_lineage(self, chart_list, dashboard_name):
        metadata = OpenMetadata(self.metadata_config)
        for chart in chart_list:
            try:
                chart_details = chart["card"]
                resp_tables = self.req_get(f"/api/table/{chart_details['table_id']}")
                if resp_tables.status_code == 200:
                    table = resp_tables.json()
                    table_fqdn = f"{self.config.database_service_name}.{table['schema']}.{table['name']}"
                    dashboard_fqdn = (
                        f"{self.dashboard_service.name}.{quote(dashboard_name)}"
                    )
                    table_entity = metadata.get_by_name(entity=Table, fqdn=table_fqdn)
                    chart_entity = metadata.get_by_name(
                        entity=Model_Dashboard, fqdn=dashboard_fqdn
                    )
                    logger.debug("from entity {}".format(table_entity))
                    lineage = AddLineage(
                        edge=EntitiesEdge(
                            fromEntity=EntityReference(
                                id=table_entity.id.__root__, type="table"
                            ),
                            toEntity=EntityReference(
                                id=chart_entity.id.__root__, type="dashboard"
                            ),
                        )
                    )
                    yield lineage
            except Exception as err:
                logger.error(traceback.print_exc())

    def req_get(self, path):
        return requests.get(self.config.host_port + path, headers=self.metabase_session)

    def get_status(self) -> SourceStatus:
        return self.status

    def close(self):
        pass

    def prepare(self):
        pass



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import collections
from typing import Callable, Iterable, Tuple, TypeVar

T = TypeVar("T")
K = TypeVar("K")


def group_by(
    iterable: Iterable[T], key: Callable[[T], K]
) -> Iterable[Tuple[K, Iterable[T]]]:

    values = collections.defaultdict(list)
    for v in iterable:
        values[key(v)].append(v)
    return values.items()



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc
import json

NODE_KEY = "KEY"
NODE_LABEL = "LABEL"
NODE_REQUIRED_HEADERS = {NODE_LABEL, NODE_KEY}


class JsonSerializable(object, metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        pass

    @staticmethod
    def snake_to_camel(s):
        # change property names like first_name to firstName
        # property names that are already in camelCase like isBot are unchanged
        a = s.split("_")
        if len(a) > 1:
            a[0] = a[0].lower()
            a[1:] = [u.title() for u in a[1:]]
        return "".join(a)

    @staticmethod
    def serialize(obj):
        return {JsonSerializable.snake_to_camel(k): v for k, v in obj.__dict__.items()}

    def to_json(self):
        return json.dumps(
            JsonSerializable.serialize(self),
            indent=4,
            default=JsonSerializable.serialize,
        )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

from metadata.config.common import ConfigModel


@dataclass  # type: ignore[misc]
class AuthenticationProvider(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def create(cls, config: ConfigModel) -> "AuthenticationProvider":
        pass

    @abstractmethod
    def auth_token(self) -> str:
        pass



import json
import traceback
from typing import TYPE_CHECKING, Dict, List, Optional

import dateutil
from airflow.configuration import conf
from airflow.lineage.backend import LineageBackend

from metadata.generated.schema.api.data.createPipeline import (
    CreatePipelineEntityRequest,
)
from metadata.generated.schema.api.lineage.addLineage import AddLineage
from metadata.generated.schema.api.services.createPipelineService import (
    CreatePipelineServiceEntityRequest,
)
from metadata.generated.schema.entity.data.pipeline import Task
from metadata.generated.schema.entity.data.table import Table
from metadata.generated.schema.entity.services.pipelineService import (
    PipelineService,
    PipelineServiceType,
)
from metadata.generated.schema.type.entityLineage import EntitiesEdge
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.utils.helpers import convert_epoch_to_iso

if TYPE_CHECKING:
    from airflow import DAG
    from airflow.models.baseoperator import BaseOperator

from metadata.config.common import ConfigModel


class OpenMetadataLineageConfig(ConfigModel):
    airflow_service_name: str = "airflow"
    api_endpoint: str = "http://localhost:8585"
    auth_provider_type: str = "no-auth"
    secret_key: str = None


def get_lineage_config() -> OpenMetadataLineageConfig:
    """Load the lineage config from airflow_provider_openmetadata.cfg."""
    airflow_service_name = conf.get(
        "lineage", "airflow_service_name", fallback="airflow"
    )
    api_endpoint = conf.get(
        "lineage", "openmetadata_api_endpoint", fallback="http://localhost:8585"
    )
    auth_provider_type = conf.get("lineage", "auth_provider_type", fallback="no-auth")
    secret_key = conf.get("lineage", "secret_key", fallback=None)
    return OpenMetadataLineageConfig.parse_obj(
        {
            "airflow_service_name": airflow_service_name,
            "api_endpoint": api_endpoint,
            "auth_provider_type": auth_provider_type,
            "secret_key": secret_key,
        }
    )


allowed_task_keys = [
    "_downstream_task_ids",
    "_inlets",
    "_outlets",
    "_task_type",
    "_task_module",
    "depends_on_past",
    "email",
    "label",
    "execution_timeout",
    "end_date",
    "start_date",
    "sla",
    "sql",
    "task_id",
    "trigger_rule",
    "wait_for_downstream",
]
allowed_flow_keys = [
    "_access_control",
    "_concurrency",
    "_default_view",
    "catchup",
    "fileloc",
    "is_paused_upon_creation",
    "start_date",
    "tags",
    "timezone",
]


def parse_lineage_to_openmetadata(
    config: OpenMetadataLineageConfig,
    context: Dict,
    operator: "BaseOperator",
    inlets: List,
    outlets: List,
    client: OpenMetadata,
):
    import ast

    from airflow.serialization.serialized_objects import (
        SerializedBaseOperator,
        SerializedDAG,
    )

    operator.log.info("Parsing Lineage for OpenMetadata")
    dag: "DAG" = context["dag"]
    task: BaseOperator = context["task"]

    pipeline_service_url = conf.get("webserver", "base_url")
    dag_url = f"{pipeline_service_url}/tree?dag_id={dag.dag_id}"
    task_url = f"{pipeline_service_url}/taskinstance/list/?flt1_dag_id_equals={dag.dag_id}&_flt_3_task_id={task.task_id}"

    dag_properties: Dict[str, str] = {
        key: repr(value) for (key, value) in SerializedDAG.serialize_dag(dag).items()
    }
    for key in dag.get_serialized_fields():
        if key not in dag_properties:
            dag_properties[key] = repr(getattr(dag, key))
    task_properties: Dict[str, str] = {
        key: repr(value)
        for (key, value) in SerializedBaseOperator.serialize_operator(task).items()
    }
    for key in task.get_serialized_fields():
        if key not in task_properties:
            task_properties[key] = repr(getattr(task, key))

    task_properties = {
        k: v for (k, v) in task_properties.items() if k in allowed_task_keys
    }

    dag_properties = {
        k: v for (k, v) in dag_properties.items() if k in allowed_flow_keys
    }

    operator.log.info("Task Properties {}".format(task_properties))
    operator.log.info("DAG properties {}".format(dag_properties))
    # operator.log.info("Pipeline Context {}".format(context))

    timestamp = int(dateutil.parser.parse(context["ts"]).timestamp() * 1000)
    owner = dag.owner
    tags = dag.tags
    airflow_service_entity = None
    operator.log.info("Get Airflow Service ID")
    airflow_service_entity = client.get_by_name(
        entity=PipelineService, fqdn=config.airflow_service_name
    )
    if airflow_service_entity is None:
        pipeline_service = CreatePipelineServiceEntityRequest(
            name=config.airflow_service_name,
            serviceType=PipelineServiceType.Airflow,
            pipelineUrl=pipeline_service_url,
        )
        airflow_service_entity = client.create_or_update(pipeline_service)

    operator.log.info("airflow service entity {}", airflow_service_entity)
    operator.log.info(task_properties)
    operator.log.info(dag_properties)
    downstream_tasks = []
    dag_start_date = convert_epoch_to_iso(int(float(dag_properties["start_date"])))

    if "_downstream_task_ids" in task_properties:
        downstream_tasks = ast.literal_eval(task_properties["_downstream_task_ids"])

    operator.log.info("downstream tasks {}".format(downstream_tasks))
    task_start_date = (
        task_properties["start_date"].isoformat()
        if "start_time" in task_properties
        else None
    )
    task_end_date = (
        task_properties["end_date"].isoformat()
        if "end_time" in task_properties
        else None
    )

    task = Task(
        name=task_properties["task_id"],
        displayName=task_properties["label"],
        taskUrl=task_url,
        taskType=task_properties["_task_type"],
        startDate=task_start_date,
        endDate=task_end_date,
        downstreamTasks=downstream_tasks,
    )
    create_pipeline = CreatePipelineEntityRequest(
        name=dag.dag_id,
        displayName=dag.dag_id,
        description=dag.description,
        pipelineUrl=dag_url,
        startDate=dag_start_date,
        tasks=[task],
        service=EntityReference(id=airflow_service_entity.id, type="pipelineService"),
    )
    pipeline = client.create_or_update(create_pipeline)
    operator.log.info("Create Pipeline {}".format(pipeline))

    operator.log.info("Parsing Lineage")
    for table in inlets:
        table_entity = client.get_by_name(entity=Table, fqdn=table.fullyQualifiedName)
        operator.log.debug("from entity {}".format(table_entity))
        lineage = AddLineage(
            edge=EntitiesEdge(
                fromEntity=EntityReference(id=table_entity.id, type="table"),
                toEntity=EntityReference(id=pipeline.id, type="pipeline"),
            )
        )
        operator.log.debug("from lineage {}".format(lineage))
        client.add_lineage(lineage)

    for table in outlets:
        table_entity = client.get_by_name(entity=Table, fqdn=table.fullyQualifiedName)
        operator.log.debug("to entity {}".format(table_entity))
        lineage = AddLineage(
            edge=EntitiesEdge(
                fromEntity=EntityReference(id=pipeline.id, type="pipeline"),
                toEntity=EntityReference(id=table_entity.id, type="table"),
            )
        )
        operator.log.debug("to lineage {}".format(lineage))
        client.add_lineage(lineage)


def is_airflow_version_1() -> bool:
    try:
        from airflow.hooks.base import BaseHook

        return False
    except ModuleNotFoundError:
        from airflow.hooks.base_hook import BaseHook

        return True


class OpenMetadataLineageBackend(LineageBackend):
    """
    Sends lineage data from tasks to OpenMetadata.
    Configurable via ``airflow_provider_openmetadata.cfg`` as follows: ::
    [lineage]
    backend = airflow_provider_openmetadata.lineage.OpenMetadataLineageBackend
    airflow_service_name = airflow #make sure this service_name matches the one configured in openMetadata
    openmetadata_api_endpoint = http://localhost:8585
    auth_provider_type = no-auth # use google here if you are configuring google as SSO
    secret_key = google-client-secret-key # it needs to be configured only if you are using google as SSO
    """

    def __init__(self) -> None:
        super().__init__()
        _ = get_lineage_config()

    @staticmethod
    def send_lineage(
        operator: "BaseOperator",
        inlets: Optional[List] = None,
        outlets: Optional[List] = None,
        context: Dict = None,
    ) -> None:

        try:
            config = get_lineage_config()
            metadata_config = MetadataServerConfig.parse_obj(
                {
                    "api_endpoint": config.api_endpoint,
                    "auth_provider_type": config.auth_provider_type,
                    "secret_key": config.secret_key,
                }
            )
            client = OpenMetadata(metadata_config)
            op_inlets = []
            op_outlets = []
            if (
                isinstance(operator._inlets, list)
                and len(operator._inlets) == 1
                and isinstance(operator._inlets[0], dict)
                and not is_airflow_version_1()
            ):
                op_inlets = operator._inlets[0].get("tables", [])

            if (
                isinstance(operator._outlets, list)
                and len(operator._outlets) == 1
                and isinstance(operator._outlets[0], dict)
                and not is_airflow_version_1()
            ):
                op_outlets = operator._outlets[0].get("tables", [])
            if len(op_inlets) == 0 and len(operator.inlets) != 0:
                op_inlets = operator.inlets
            if len(op_outlets) == 0 and len(operator.outlets) != 0:
                op_outlets = operator.outlets

            parse_lineage_to_openmetadata(
                config, context, operator, op_inlets, op_outlets, client
            )
        except Exception as e:
            operator.log.error(traceback.format_exc())
            operator.log.error(e)



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import random
import uuid
from dataclasses import dataclass, field
from typing import Iterable, List

from faker import Faker

from metadata.generated.schema.api.data.createTopic import CreateTopicEntityRequest
from metadata.generated.schema.api.services.createDashboardService import (
    CreateDashboardServiceEntityRequest,
)
from metadata.generated.schema.api.services.createDatabaseService import (
    CreateDatabaseServiceEntityRequest,
)
from metadata.generated.schema.api.services.createMessagingService import (
    CreateMessagingServiceEntityRequest,
)
from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.table import Column, Constraint, Table
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.generated.schema.type.tagLabel import TagLabel
from metadata.ingestion.api.common import Entity
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from metadata.ingestion.models.table_metadata import Chart, Dashboard
from metadata.ingestion.ometa.client import APIError
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLConnectionConfig
from metadata.utils.helpers import snake_to_camel

logger: logging.Logger = logging.getLogger(__name__)


class SampleEntitySourceConfig(SQLConnectionConfig):
    no_of_services: int
    no_of_databases: int
    no_of_tables: int
    no_of_columns: int
    no_of_dashboards: int
    no_of_charts: int
    no_of_topics: int
    generate_tables: bool = True
    generate_dashboards: bool = True
    generate_topics: bool = True

    def get_connection_url(self):
        pass


@dataclass
class SampleEntitySourceStatus(SourceStatus):
    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)

    def scanned(self, entity_type: str, entity_name: str) -> None:
        self.success.append(entity_name)
        logger.info("{} Scanned: {}".format(entity_type, entity_name))

    def filtered(self, entity_type: str, entity_name: str, err: str) -> None:
        self.warnings.append(entity_name)
        logger.warning("Dropped {} {} due to {}".format(entity_type, entity_name, err))


class SampleEntitySource(Source[Entity]):
    def __init__(
        self,
        config: SampleEntitySourceConfig,
        metadata_config: MetadataServerConfig,
        ctx,
    ):
        super().__init__(ctx)
        self.faker = Faker()
        self.status = SampleEntitySourceStatus()
        self.config = config
        self.metadata_config = metadata_config
        self.metadata = OpenMetadata(metadata_config)
        self.service_name = lambda: self.faker.word()
        self.service_type = lambda: random.choice(
            ["BigQuery", "Hive", "MSSQL", "MySQL", "Postgres", "Redshift", "Snowflake"]
        )
        self.database_name = lambda: self.faker.word()
        self.table_name = lambda: self.faker.word()
        self.column_name = lambda: self.faker.word()
        self.description = lambda: self.faker.text()
        self.chart_ids = lambda: self.faker.random_int()
        self.tags = self.__get_tags()
        self.tagFQN = lambda: self.faker.first_name()
        self.labelType = lambda: random.choice(
            ["Automated", "Derived", "Manual", "Propagated"]
        )
        self.state = lambda: random.choice(["Confirmed", "Suggested"])
        self.href = lambda: self.faker.url()
        self.col_type = lambda: random.choice(["INT", "STRING", "VARCHAR", "DATE"])
        self.chart_type = lambda: random.choice(
            ["Area", "Line", "Table", "Bar", "Pie", "Histogram", "Scatter", "Text"]
        )
        self.col_constraint = lambda: random.choice(
            [Constraint.UNIQUE, Constraint.NOT_NULL, Constraint.NULL]
        )

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = SampleEntitySourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def __get_tags(self) -> {}:
        return self.metadata.list_tags_by_category("user")

    def scan(self, text):
        types = set()
        for pii_type in self.regex:
            if self.regex[pii_type].match(text) is not None:
                types.add(pii_type.name)

        logging.debug("PiiTypes are %s", ",".join(str(x) for x in list(types)))
        return list(types)

    def next_record(self) -> Iterable[Entity]:
        if self.config.generate_tables:
            yield from self.ingest_tables()
        if self.config.generate_dashboards:
            yield from self.ingest_dashboards()
        if self.config.generate_topics:
            yield from self.ingest_topics()

    def ingest_tables(self) -> Iterable[OMetaDatabaseAndTable]:
        for h in range(self.config.no_of_services):
            service = {
                "jdbc": {"connectionUrl": f"jdbc://localhost", "driverClass": "jdbc"},
                "name": self.service_name(),
                "description": self.description(),
                "serviceType": self.service_type(),
            }
            create_service = None
            while True:
                try:
                    create_service = self.metadata.create_or_update(
                        CreateDatabaseServiceEntityRequest(**service)
                    )
                    break
                except APIError as err:
                    continue

            logger.info(
                "Ingesting service {}/{}".format(h + 1, self.config.no_of_services)
            )
            for i in range(self.config.no_of_databases):
                db = Database(
                    id=uuid.uuid4(),
                    name=self.database_name().replace(".", "_"),
                    description=self.description(),
                    service=EntityReference(
                        id=create_service.id, type=self.config.service_type
                    ),
                )

                logger.info(
                    "Ingesting database {}/{} in service: {}/{}".format(
                        i + 1,
                        self.config.no_of_databases,
                        h + 1,
                        self.config.no_of_services,
                    )
                )

                for j in range(self.config.no_of_tables):
                    table_columns = []
                    table_entity = Table(
                        id=uuid.uuid4(),
                        name=self.table_name().replace(".", "_"),
                        tableType="Regular",
                        description=self.description(),
                        columns=table_columns,
                    )
                    row_order = 0
                    for t in range(self.config.no_of_columns):
                        tag_labels = []
                        tag_entity = random.choice(self.tags)
                        tag_labels.append(
                            TagLabel(
                                tagFQN=tag_entity.fullyQualifiedName,
                                labelType="Automated",
                                state="Suggested",
                                href=tag_entity.href,
                            )
                        )
                        table_columns.append(
                            Column(
                                name=self.column_name(),
                                description=self.description(),
                                dataType=self.col_type(),
                                constraint=self.col_constraint(),
                                dataLength=100,
                                ordinalPosition=row_order,
                                tags=tag_labels,
                            )
                        )
                        table_entity.columns = table_columns
                        row_order = row_order + 1

                    table_and_db = OMetaDatabaseAndTable(
                        table=table_entity, database=db
                    )
                    yield table_and_db

    def ingest_dashboards(self) -> Iterable[Dashboard]:
        for h in range(self.config.no_of_services):
            create_service = None
            while True:
                try:
                    service = {
                        "name": self.service_name(),
                        "description": self.description(),
                        "dashboardUrl": "http://localhost:8088",
                        "userName": "admin",
                        "password": "admin",
                        "serviceType": "Superset",
                    }
                    create_service = self.metadata.create_or_update(
                        CreateDashboardServiceEntityRequest(**service)
                    )
                    break
                except APIError as err:
                    continue

            logger.info(
                "Ingesting service {}/{}".format(h + 1, self.config.no_of_services)
            )
            for i in range(self.config.no_of_dashboards):
                logger.info(
                    "Ingesting dashboard {}/{} in service: {}/{}".format(
                        i + 1,
                        self.config.no_of_databases,
                        h + 1,
                        self.config.no_of_services,
                    )
                )
                chart_ids = []
                for j in range(self.config.no_of_charts):
                    charts = []
                    chart_id = self.chart_ids()
                    chart_entity = Chart(
                        id=uuid.uuid4(),
                        name=str(chart_id),
                        displayName=self.table_name(),
                        description=self.description(),
                        chart_type=self.chart_type(),
                        chartId=str(chart_id),
                        url="http://superset:8080/chartUrl",
                        service=EntityReference(
                            id=create_service.id, type="dashboardService"
                        ),
                    )
                    chart_ids.append(str(chart_id))
                    yield chart_entity

                dashboard = Dashboard(
                    id=uuid.uuid4(),
                    name=str(self.chart_ids()),
                    displayName=self.table_name(),
                    description=self.description(),
                    url="http://superset:8080/dashboardUrl",
                    charts=chart_ids,
                    service=EntityReference(
                        id=create_service.id, type="dashboardService"
                    ),
                )
                yield dashboard

    def ingest_topics(self) -> Iterable[CreateTopicEntityRequest]:
        for h in range(self.config.no_of_services):
            create_service = None
            while True:
                try:
                    service = {
                        "name": self.service_name(),
                        "description": self.description(),
                        "brokers": ["localhost:9092"],
                        "schemaRegistry": "http://localhost:8081",
                        "serviceType": "Kafka",
                    }
                    create_service = self.metadata.create_or_update(
                        CreateMessagingServiceEntityRequest(**service)
                    )
                    break
                except APIError as err:
                    continue

            logger.info(
                "Ingesting service {}/{}".format(h + 1, self.config.no_of_services)
            )
            for j in range(self.config.no_of_topics):
                topic_entity = CreateTopicEntityRequest(
                    name=self.table_name(),
                    description=self.description(),
                    partitions=self.chart_ids(),
                    retentionSize=322122382273,
                    replicationFactor=2,
                    maximumMessageSize=167,
                    cleanupPolicies=["delete"],
                    schemaType="Avro",
                    schemaText='{"namespace":"org.open-metadata.kafka","name":"Customer","type":"record","fields":[{"name":"id","type":"string"},{"name":"first_name","type":"string"},{"name":"last_name","type":"string"},{"name":"email","type":"string"},{"name":"address_line_1","type":"string"},{"name":"address_line_2","type":"string"},{"name":"post_code","type":"string"},{"name":"country","type":"string"}]}',
                    service=EntityReference(
                        id=create_service.id, type="messagingService"
                    ),
                )
                yield topic_entity

    def close(self):
        pass

    def get_status(self):
        return self.status



# generated by datamodel-codegen:
#   filename:  schema/api/data/createTopic.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field, conint

from ...entity.data import topic
from ...type import entityReference, tagLabel


class CreateTopicEntityRequest(BaseModel):
    name: topic.TopicName = Field(
        ..., description='Name that identifies this topic instance uniquely.'
    )
    description: Optional[str] = Field(
        None,
        description='Description of the topic instance. What it has and how to use it.',
    )
    service: entityReference.EntityReference = Field(
        ..., description='Link to the messaging service where this topic is hosted in'
    )
    partitions: conint(ge=1) = Field(
        ..., description='Number of partitions into which the topic is divided.'
    )
    schemaText: Optional[str] = Field(
        None,
        description='Schema used for message serialization. Optional as some topics may not have associated schemas.',
    )
    schemaType: Optional[topic.SchemaType] = Field(
        None, description='Schema used for message serialization.'
    )
    cleanupPolicies: Optional[List[topic.CleanupPolicy]] = Field(
        None,
        description='Topic clean up policy. For Kafka - `cleanup.policy` configuration.',
    )
    replicationFactor: Optional[int] = Field(
        None, description='Replication Factor in integer (more than 1).'
    )
    retentionTime: Optional[float] = Field(
        None,
        description='Retention time in milliseconds. For Kafka - `retention.ms` configuration.',
    )
    maximumMessageSize: Optional[int] = Field(
        None,
        description='Maximum message size in bytes. For Kafka - `max.message.bytes` configuration.',
    )
    minimumInSyncReplicas: Optional[int] = Field(
        None,
        description='Minimum number replicas in sync to control durability. For Kafka - `min.insync.replicas` configuration.',
    )
    retentionSize: Optional[float] = Field(
        '-1',
        description='Maximum size of a partition in bytes before old data is discarded. For Kafka - `retention.bytes` configuration.',
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this topic'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this topic'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Generic, List

from metadata.ingestion.api.closeable import Closeable
from metadata.ingestion.api.common import Entity, WorkflowContext
from metadata.ingestion.api.status import Status


@dataclass
class ProcessorStatus(Status):
    records = 0
    warnings: List[Any] = field(default_factory=list)
    failures: List[Any] = field(default_factory=list)

    def processed(self, record: Any):
        self.records += 1

    def warning(self, info: Any) -> None:
        self.warnings.append(info)

    def failure(self, info: Any) -> None:
        self.failures.append(info)


@dataclass
class Processor(Closeable, Generic[Entity], metaclass=ABCMeta):
    ctx: WorkflowContext

    @classmethod
    @abstractmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ) -> "Processor":
        pass

    @abstractmethod
    def process(self, record: Entity) -> Entity:
        pass

    @abstractmethod
    def get_status(self) -> ProcessorStatus:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from metadata.config.common import ConfigModel
from metadata.generated.schema.entity.data.dashboard import Dashboard
from metadata.generated.schema.entity.data.pipeline import Pipeline
from metadata.generated.schema.entity.data.table import Table
from metadata.generated.schema.entity.data.topic import Topic
from metadata.ingestion.api.common import Entity, WorkflowContext
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig

logger = logging.getLogger(__name__)


class MetadataTablesRestSourceConfig(ConfigModel):
    include_tables: Optional[bool] = True
    include_topics: Optional[bool] = True
    include_dashboards: Optional[bool] = True
    include_pipelines: Optional[bool] = True
    limit_records: int = 1000


@dataclass
class MetadataSourceStatus(SourceStatus):

    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)

    def scanned_table(self, table_name: str) -> None:
        self.success.append(table_name)
        logger.info("Table Scanned: {}".format(table_name))

    def scanned_topic(self, topic_name: str) -> None:
        self.success.append(topic_name)
        logger.info("Topic Scanned: {}".format(topic_name))

    def scanned_dashboard(self, dashboard_name: str) -> None:
        self.success.append(dashboard_name)
        logger.info("Dashboard Scanned: {}".format(dashboard_name))

    def filtered(
        self, table_name: str, err: str, dataset_name: str = None, col_type: str = None
    ) -> None:
        self.warnings.append(table_name)
        logger.warning("Dropped Entity {} due to {}".format(table_name, err))


class MetadataSource(Source[Entity]):
    config: MetadataTablesRestSourceConfig
    report: SourceStatus

    def __init__(
        self,
        config: MetadataTablesRestSourceConfig,
        metadata_config: MetadataServerConfig,
        ctx: WorkflowContext,
    ):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.status = MetadataSourceStatus()
        self.wrote_something = False
        self.metadata = OpenMetadata(self.metadata_config)
        self.tables = None
        self.topics = None

    def prepare(self):
        pass

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = MetadataTablesRestSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def next_record(self) -> Iterable[Entity]:
        yield from self.fetch_table()
        yield from self.fetch_topic()
        yield from self.fetch_dashboard()
        yield from self.fetch_pipeline()

    def fetch_table(self) -> Table:
        if self.config.include_tables:
            after = None
            while True:
                table_entities = self.metadata.list_entities(
                    entity=Table,
                    fields=[
                        "columns",
                        "tableConstraints",
                        "usageSummary",
                        "owner",
                        "tags",
                        "followers",
                    ],
                    after=after,
                    limit=self.config.limit_records,
                )
                for table in table_entities.entities:
                    self.status.scanned_table(table.name.__root__)
                    yield table
                if table_entities.after is None:
                    break
                after = table_entities.after

    def fetch_topic(self) -> Topic:
        if self.config.include_topics:
            after = None
            while True:
                topic_entities = self.metadata.list_entities(
                    entity=Topic,
                    fields=["owner", "tags", "followers"],
                    after=after,
                    limit=self.config.limit_records,
                )
                for topic in topic_entities.entities:
                    self.status.scanned_topic(topic.name.__root__)
                    yield topic
                if topic_entities.after is None:
                    break
                after = topic_entities.after

    def fetch_dashboard(self) -> Dashboard:
        if self.config.include_dashboards:
            after = None
            while True:
                dashboard_entities = self.metadata.list_entities(
                    entity=Dashboard,
                    fields=[
                        "owner",
                        "tags",
                        "followers",
                        "charts",
                        "usageSummary",
                    ],
                    after=after,
                    limit=self.config.limit_records,
                )
                for dashboard in dashboard_entities.entities:
                    self.status.scanned_dashboard(dashboard.name)
                    yield dashboard
                if dashboard_entities.after is None:
                    break
                after = dashboard_entities.after

    def fetch_pipeline(self) -> Pipeline:
        if self.config.include_pipelines:
            after = None
            while True:
                pipeline_entities = self.metadata.list_entities(
                    entity=Pipeline,
                    fields=["owner", "tags", "followers", "tasks"],
                    after=after,
                    limit=self.config.limit_records,
                )
                for pipeline in pipeline_entities.entities:
                    self.status.scanned_dashboard(pipeline.name)
                    yield pipeline
                if pipeline_entities.after is None:
                    break
                after = pipeline_entities.after

    def get_status(self) -> SourceStatus:
        return self.status

    def close(self):
        pass



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import re
import textwrap
import traceback
import uuid
from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from pydantic import SecretStr

from metadata.config.common import ConfigModel
from metadata.generated.schema.api.services.createDatabaseService import (
    CreateDatabaseServiceEntityRequest,
)
from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.table import Column, Table
from metadata.generated.schema.entity.services.dashboardService import (
    DashboardServiceType,
)
from metadata.generated.schema.entity.services.databaseService import DatabaseService
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import Entity
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from metadata.ingestion.models.table_metadata import Chart, Dashboard
from metadata.ingestion.models.user import User
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.neo4j_helper import Neo4JConfig, Neo4jHelper
from metadata.utils.column_helpers import check_column_complex_type, get_column_type
from metadata.utils.helpers import get_dashboard_service_or_create

logger: logging.Logger = logging.getLogger(__name__)


class AmundsenConfig(ConfigModel):
    neo4j_username: Optional[str] = None
    neo4j_password: Optional[SecretStr] = None
    neo4j_url: str
    neo4j_max_connection_life_time: int = 50
    neo4j_encrypted: bool = True
    neo4j_validate_ssl: bool = False


PRIMITIVE_TYPES = ["int", "char", "varchar"]

NEO4J_AMUNDSEN_TABLE_QUERY = textwrap.dedent(
    """
        MATCH (db:Database)<-[:CLUSTER_OF]-(cluster:Cluster)
        <-[:SCHEMA_OF]-(schema:Schema)<-[:TABLE_OF]-(table:Table)
        OPTIONAL MATCH (table)-[:DESCRIPTION]->(table_description:Description)
        OPTIONAL MATCH (schema)-[:DESCRIPTION]->(schema_description:Description)
        OPTIONAL MATCH (table)-[:DESCRIPTION]->(prog_descs:Programmatic_Description)
        WITH db, cluster, schema, schema_description, table, table_description,
        COLLECT(prog_descs.description) as programmatic_descriptions
        OPTIONAL MATCH (table)-[:TAGGED_BY]->(tags:Tag) WHERE tags.tag_type='default'
        WITH db, cluster, schema, schema_description, table, table_description, programmatic_descriptions,
        COLLECT(DISTINCT tags.key) as tags
        OPTIONAL MATCH (table)-[:HAS_BADGE]->(badges:Badge)
        WITH db, cluster, schema, schema_description, table, table_description, programmatic_descriptions, tags,
        COLLECT(DISTINCT badges.key) as badges
        OPTIONAL MATCH (table)-[read:READ_BY]->(user:User)
        WITH db, cluster, schema, schema_description, table, table_description, programmatic_descriptions, tags, badges,
        SUM(read.read_count) AS total_usage,
        COUNT(DISTINCT user.email) as unique_usage
        OPTIONAL MATCH (table)-[:COLUMN]->(col:Column)
        OPTIONAL MATCH (col)-[:DESCRIPTION]->(col_description:Description)
        WITH db, cluster, schema, schema_description, table, table_description, tags, badges, total_usage, unique_usage,
        programmatic_descriptions,
        COLLECT(col.name) AS column_names, COLLECT(col_description.description) AS column_descriptions,
        COLLECT(col.col_type) AS column_types
        OPTIONAL MATCH (table)-[:LAST_UPDATED_AT]->(time_stamp:Timestamp)
        RETURN db.name as database, cluster.name AS cluster, schema.name AS schema,
        schema_description.description AS schema_description,
        table.name AS name, table.key AS key, table_description.description AS description,
        time_stamp.last_updated_timestamp AS last_updated_timestamp,
        column_names,
        column_descriptions,
        column_types,
        total_usage,
        unique_usage,
        tags,
        badges,
        programmatic_descriptions
        ORDER BY table.name;
        """
)

NEO4J_AMUNDSEN_USER_QUERY = textwrap.dedent(
    """
        MATCH (user:User)
        OPTIONAL MATCH (user)-[read:READ]->(a)
        OPTIONAL MATCH (user)-[own:OWNER_OF]->(b)
        OPTIONAL MATCH (user)-[follow:FOLLOWED_BY]->(c)
        OPTIONAL MATCH (user)-[manage_by:MANAGE_BY]->(manager)
        with user, a, b, c, read, own, follow, manager
        where user.full_name is not null
        return user.email as email, user.first_name as first_name, user.last_name as last_name,
        user.full_name as full_name, user.github_username as github_username, user.team_name as team_name,
        user.employee_type as employee_type, manager.email as manager_email,
        user.slack_id as slack_id, user.is_active as is_active, user.role_name as role_name,
        REDUCE(sum_r = 0, r in COLLECT(DISTINCT read)| sum_r + r.read_count) AS total_read,
        count(distinct b) as total_own,
        count(distinct c) AS total_follow
        order by user.email
        """
)

NEO4J_AMUNDSEN_DASHBOARD_QUERY = textwrap.dedent(
    """
         MATCH (dashboard:Dashboard)
         MATCH (dashboard)-[:DASHBOARD_OF]->(dbg:Dashboardgroup)
         MATCH (dbg)-[:DASHBOARD_GROUP_OF]->(cluster:Cluster)
         OPTIONAL MATCH (dashboard)-[:DESCRIPTION]->(db_descr:Description)
         OPTIONAL MATCH (dbg)-[:DESCRIPTION]->(dbg_descr:Description)
         OPTIONAL MATCH (dashboard)-[:EXECUTED]->(last_exec:Execution)
         WHERE split(last_exec.key, '/')[5] = '_last_successful_execution'
         OPTIONAL MATCH (dashboard)-[read:READ_BY]->(user:User)
         WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, SUM(read.read_count) AS total_usage
         OPTIONAL MATCH (dashboard)-[:HAS_QUERY]->(query:Query)-[:HAS_CHART]->(chart:Chart)
         WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, COLLECT(DISTINCT query.name) as query_names,
         COLLECT(chart.name) as chart_names,
         COLLECT(chart.id) as chart_ids,
         COLLECT(chart.url) as chart_urls,
         COLLECT(chart.type) as chart_types,
         total_usage
         OPTIONAL MATCH (dashboard)-[:TAGGED_BY]->(tags:Tag) WHERE tags.tag_type='default'
         WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, chart_names, chart_ids, chart_urls,
         chart_types, total_usage,
         COLLECT(DISTINCT tags.key) as tags
         OPTIONAL MATCH (dashboard)-[:HAS_BADGE]->(badges:Badge)
         WITH  dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, chart_names, chart_ids, chart_urls,
         chart_types, total_usage, tags,
         COLLECT(DISTINCT badges.key) as badges
         RETURN dbg.name as group_name, dashboard.name as name, cluster.name as cluster,
         coalesce(db_descr.description, '') as description,
         coalesce(dbg.description, '') as group_description, dbg.dashboard_group_url as group_url,
         dashboard.dashboard_id as dashboard_id,
         dashboard.dashboard_url as url, dashboard.key as uri,
         split(dashboard.key, '_')[0] as product, toInteger(last_exec.timestamp) as last_successful_run_timestamp,
         query_names, chart_names, chart_ids, chart_urls, chart_types, total_usage, tags, badges
         order by dbg.name
        """
)


@dataclass
class AmundsenStatus(SourceStatus):
    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    filtered: List[str] = field(default_factory=list)

    def scanned(self, entity_name: str) -> None:
        self.success.append(entity_name)
        logger.info("Entity Scanned: {}".format(entity_name))

    def failure(self, key: str, reason: str) -> None:
        self.failures.append({key: reason})


class AmundsenSource(Source[Entity]):
    def __init__(
        self, config: AmundsenConfig, metadata_config: MetadataServerConfig, ctx
    ):
        self.config = config
        self.metadata_config = metadata_config
        self.ctx = ctx
        neo4j_config = Neo4JConfig(
            username=self.config.neo4j_username,
            password=self.config.neo4j_password.get_secret_value(),
            neo4j_url=self.config.neo4j_url,
            max_connection_life_time=self.config.neo4j_max_connection_life_time,
            neo4j_encrypted=self.config.neo4j_encrypted,
            neo4j_validate_ssl=self.config.neo4j_validate_ssl,
        )
        self.neo4j_helper = Neo4jHelper(neo4j_config)
        self.status = AmundsenStatus()

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = AmundsenConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def next_record(self) -> Iterable[Entity]:
        user_entities = self.neo4j_helper.execute_query(NEO4J_AMUNDSEN_USER_QUERY)
        for user in user_entities:
            yield from self.create_user_entity(user)

        table_entities = self.neo4j_helper.execute_query(NEO4J_AMUNDSEN_TABLE_QUERY)
        for table in table_entities:
            yield from self.create_table_entity(table)

        dashboard_entities = self.neo4j_helper.execute_query(
            NEO4J_AMUNDSEN_DASHBOARD_QUERY
        )
        for dashboard in dashboard_entities:
            yield from self.create_chart_entity(dashboard)
            yield from self.create_dashboard_entity(dashboard)

    def create_user_entity(self, user):
        try:
            user_metadata = User(
                email=user["email"],
                first_name=user["first_name"],
                last_name=user["last_name"],
                name=user["full_name"],
                team_name=user["team_name"],
                is_active=user["is_active"],
            )
            self.status.scanned(user_metadata.email)
            yield user_metadata
        except Exception as err:
            logger.error(err)

    def create_table_entity(self, table):
        try:
            service_name = table["cluster"]
            service_type = table["database"]
            service_entity = self.get_database_service_or_create(
                service_name, service_type
            )
            database = Database(
                name=table["schema"],
                service=EntityReference(id=service_entity.id, type=service_type),
            )
            columns: List[Column] = []
            row_order = 1
            for (name, description, data_type) in zip(
                table["column_names"],
                table["column_descriptions"],
                table["column_types"],
            ):
                # Amundsen merges the length into type itself. Instead of making changes to our generic type builder
                # we will do a type match and see if it matches any primitive types and return a type
                data_type = self.get_type_primitive_type(data_type)
                (
                    col_type,
                    data_type_display,
                    arr_data_type,
                    children,
                ) = check_column_complex_type(
                    self.status, table["name"], data_type, name
                )

                col = Column(
                    name=name,
                    description=description,
                    dataType=col_type,
                    dataTypeDisplay="{}({})".format(col_type, 1)
                    if data_type_display is None
                    else f"{data_type_display}",
                    children=children,
                    arrayDataType=arr_data_type,
                    ordinalPosition=row_order,
                    dataLength=1,
                )
                row_order += 1
                columns.append(col)

            fqn = f"{service_name}.{database.name}.{table['schema']}.{table['name']}"
            table_entity = Table(
                id=uuid.uuid4(),
                name=table["name"],
                tableType="Regular",
                description=table["description"],
                fullyQualifiedName=fqn,
                columns=columns,
            )

            table_and_db = OMetaDatabaseAndTable(table=table_entity, database=database)
            self.status.scanned(table["name"])
            yield table_and_db
        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.error(f"Failed to create table entity, due to {e}")
            self.status.failure(table["name"], str(e))
            return None

    def create_dashboard_entity(self, dashboard):
        try:
            service_name = dashboard["cluster"]
            service_entity = get_dashboard_service_or_create(
                service_name,
                DashboardServiceType.Superset.name,
                "admin",
                "admin",
                "http://localhost:8088",
                self.metadata_config,
            )
            self.status.scanned(dashboard["name"])
            yield Dashboard(
                name=dashboard["name"],
                displayName=dashboard["name"],
                description="",
                url=dashboard["url"],
                charts=dashboard["chart_ids"],
                service=EntityReference(id=service_entity.id, type="dashboardService"),
            )
        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.error(f"Failed to create table entity, due to {e}")
            self.status.failure(dashboard["name"], str(e))
            return None

    def create_chart_entity(self, dashboard):
        service_name = dashboard["cluster"]
        service_entity = get_dashboard_service_or_create(
            service_name,
            DashboardServiceType.Superset.name,
            "admin",
            "admin",
            "http://localhost:8088",
            self.metadata_config,
        )

        for (name, chart_id, chart_type, url) in zip(
            dashboard["chart_names"],
            dashboard["chart_ids"],
            dashboard["chart_types"],
            dashboard["chart_urls"],
        ):
            chart = Chart(
                name=chart_id,
                displayName=name,
                description="",
                chart_url=url,
                chart_type=chart_type,
                service=EntityReference(id=service_entity.id, type="dashboardService"),
            )
            self.status.scanned(name)
            yield chart

    def close(self):
        if self.neo4j_helper is not None:
            self.neo4j_helper.close()

    def get_status(self) -> SourceStatus:
        return self.status

    def get_type_primitive_type(self, data_type):
        for p_type in PRIMITIVE_TYPES:
            if data_type.startswith(p_type):
                return p_type
        return data_type

    def get_database_service_or_create(
        self, service_name: str, service_type: str
    ) -> DatabaseService:
        metadata = OpenMetadata(self.metadata_config)
        service = metadata.get_by_name(entity=DatabaseService, fqdn=service_name)
        if service is not None:
            return service
        else:
            service = {
                "jdbc": {
                    "connectionUrl": f"jdbc://temp",
                    "driverClass": "jdbc",
                },
                "name": service_name,
                "description": "",
                "serviceType": service_type.capitalize(),
            }
            created_service = metadata.create_or_update(
                CreateDatabaseServiceEntityRequest(**service)
            )
            return created_service



# generated by datamodel-codegen:
#   filename:  schema/api/data/createLocation.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field

from ...entity.data import location
from ...type import entityReference, tagLabel


class CreateLocationEntityRequest(BaseModel):
    name: location.LocationName = Field(
        ..., description='Name that identifies this Location.'
    )
    description: Optional[str] = Field(
        None, description='Description of the location instance.'
    )
    locationType: Optional[location.LocationType] = None
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this location'
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this Location'
    )
    service: entityReference.EntityReference = Field(
        ..., description='Link to the pipeline service where this location is used'
    )



# generated by datamodel-codegen:
#   filename:  schema/api/data/createTable.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field

from ...entity.data import table
from ...type import basic, entityReference, tagLabel


class CreateTableEntityRequest(BaseModel):
    name: table.TableName = Field(
        ...,
        description='Name that identifies the this entity instance uniquely. Same as id if when name is not unique',
    )
    description: Optional[str] = Field(
        None, description='Description of entity instance.'
    )
    tableType: Optional[table.TableType] = None
    columns: List[table.Column] = Field(
        ..., description='Name of the tables in the database'
    )
    tableConstraints: Optional[List[table.TableConstraint]] = None
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this entity'
    )
    database: basic.Uuid = Field(
        ..., description='Database corresponding to this table'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this table'
    )
    viewDefinition: Optional[basic.SqlQuery] = Field(
        None, description='View Definition in SQL. Applies to TableType.View only'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import Optional, Tuple

from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLConnectionConfig, SQLSource


class BigQueryConfig(SQLConnectionConfig, SQLSource):
    scheme = "bigquery"
    project_id: Optional[str] = None
    duration: int = 1
    service_type = "BigQuery"

    def get_connection_url(self):
        if self.project_id:
            return f"{self.scheme}://{self.project_id}"
        return f"{self.scheme}://"


class BigquerySource(SQLSource):
    def __init__(self, config, metadata_config, ctx):
        super().__init__(config, metadata_config, ctx)

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config: SQLConnectionConfig = BigQueryConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = config.options[
            "credentials_path"
        ]
        return cls(config, metadata_config, ctx)

    def standardize_schema_table_names(
        self, schema: str, table: str
    ) -> Tuple[str, str]:
        segments = table.split(".")
        if len(segments) != 2:
            raise ValueError(f"expected table to contain schema name already {table}")
        if segments[0] != schema:
            raise ValueError(f"schema {schema} does not match table {table}")
        return segments[0], segments[1]

    def parse_raw_data_type(self, raw_data_type):
        return raw_data_type.replace(", ", ",").replace(" ", ":").lower()



# generated by datamodel-codegen:
#   filename:  schema/api/teams/createTeam.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field

from ...entity.teams import team
from ...type import basic, profile


class CreateTeamEntityRequest(BaseModel):
    name: team.TeamName
    displayName: Optional[str] = Field(
        None,
        description="Optional name used for display purposes. Example 'Marketing Team'",
    )
    description: Optional[str] = Field(
        None, description='Optional description of the team'
    )
    profile: Optional[profile.Profile] = Field(
        None, description='Optional team profile information'
    )
    users: Optional[List[basic.Uuid]] = Field(
        None, description='Optional IDs of users that are part of the team'
    )



# generated by datamodel-codegen:
#   filename:  schema/api/teams/createRole.json
#   timestamp: 2022-01-16T18:43:22+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field

from ...entity.teams import role


class CreateRoleEntityRequest(BaseModel):
    name: role.RoleName
    displayName: Optional[str] = Field(
        None,
        description="Optional name used for display purposes. Example 'Data Consumer'",
    )
    description: Optional[str] = Field(
        None, description='Optional description of the role'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Any, List

from .closeable import Closeable
from .common import WorkflowContext
from .status import Status


@dataclass
class BulkSinkStatus(Status):
    records = 0
    warnings: List[Any] = field(default_factory=list)
    failures: List[Any] = field(default_factory=list)

    def records_written(self, records: int):
        self.records += records

    def warning(self, info: Any) -> None:
        self.warnings.append(info)

    def failure(self, info: Any) -> None:
        self.failures.append(info)


@dataclass  # type: ignore[misc]
class BulkSink(Closeable, metaclass=ABCMeta):
    ctx: WorkflowContext

    @classmethod
    @abstractmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ) -> "BulkSink":
        pass

    @abstractmethod
    def write_records(self) -> None:
        pass

    @abstractmethod
    def get_status(self) -> BulkSinkStatus:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



# generated by datamodel-codegen:
#   filename:  schema/entity/data/table.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Extra, Field, constr

from ...type import basic, entityHistory, entityReference, tagLabel, usageDetails
from ..services import databaseService


class TableType(Enum):
    Regular = 'Regular'
    External = 'External'
    View = 'View'
    SecureView = 'SecureView'
    MaterializedView = 'MaterializedView'


class DataType(Enum):
    NUMBER = 'NUMBER'
    TINYINT = 'TINYINT'
    SMALLINT = 'SMALLINT'
    INT = 'INT'
    BIGINT = 'BIGINT'
    BYTEINT = 'BYTEINT'
    BYTES = 'BYTES'
    FLOAT = 'FLOAT'
    DOUBLE = 'DOUBLE'
    DECIMAL = 'DECIMAL'
    NUMERIC = 'NUMERIC'
    TIMESTAMP = 'TIMESTAMP'
    TIME = 'TIME'
    DATE = 'DATE'
    DATETIME = 'DATETIME'
    INTERVAL = 'INTERVAL'
    STRING = 'STRING'
    MEDIUMTEXT = 'MEDIUMTEXT'
    TEXT = 'TEXT'
    CHAR = 'CHAR'
    VARCHAR = 'VARCHAR'
    BOOLEAN = 'BOOLEAN'
    BINARY = 'BINARY'
    VARBINARY = 'VARBINARY'
    ARRAY = 'ARRAY'
    BLOB = 'BLOB'
    LONGBLOB = 'LONGBLOB'
    MEDIUMBLOB = 'MEDIUMBLOB'
    MAP = 'MAP'
    STRUCT = 'STRUCT'
    UNION = 'UNION'
    SET = 'SET'
    GEOGRAPHY = 'GEOGRAPHY'
    ENUM = 'ENUM'
    JSON = 'JSON'


class Constraint(Enum):
    NULL = 'NULL'
    NOT_NULL = 'NOT_NULL'
    UNIQUE = 'UNIQUE'
    PRIMARY_KEY = 'PRIMARY_KEY'


class ConstraintType(Enum):
    UNIQUE = 'UNIQUE'
    PRIMARY_KEY = 'PRIMARY_KEY'
    FOREIGN_KEY = 'FOREIGN_KEY'


class TableConstraint(BaseModel):
    constraintType: Optional[ConstraintType] = None
    columns: Optional[List[str]] = Field(
        None, description='List of column names corresponding to the constraint.'
    )


class ColumnName(BaseModel):
    __root__: constr(regex=r'^[^.]*$', min_length=1, max_length=128) = Field(
        ...,
        description='Local name (not fully qualified name) of the column. ColumnName is `-` when the column is not named in struct dataType. For example, BigQuery supports struct with unnamed fields.',
    )


class TableName(BaseModel):
    __root__: constr(regex=r'^[^.]*$', min_length=1, max_length=128) = Field(
        ..., description='Local name (not fully qualified name) of a table.'
    )


class FullyQualifiedColumnName(BaseModel):
    __root__: constr(min_length=1, max_length=256) = Field(
        ...,
        description='Fully qualified name of the column that includes `serviceName.databaseName.tableName.columnName[.nestedColumnName]`. When columnName is null for dataType struct fields, `field_#` where `#` is field index is used. For map dataType, for key the field name `key` is used and for the value field `value` is used.',
    )


class JoinedWithItem(BaseModel):
    fullyQualifiedName: Optional[FullyQualifiedColumnName] = None
    joinCount: Optional[int] = None


class ColumnJoins(BaseModel):
    class Config:
        extra = Extra.forbid

    columnName: Optional[ColumnName] = None
    joinedWith: Optional[List[JoinedWithItem]] = Field(
        None,
        description='Fully qualified names of the columns that this column is joined with.',
    )


class TableData(BaseModel):
    class Config:
        extra = Extra.forbid

    columns: Optional[List[ColumnName]] = Field(
        None,
        description='List of local column names (not fully qualified column names) of the table.',
    )
    rows: Optional[List[List]] = Field(
        None, description='Data for multiple rows of the table.'
    )


class ColumnProfile(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str] = Field(None, description='Column Name.')
    uniqueCount: Optional[float] = Field(
        None, description='No. of unique values in the column.'
    )
    uniqueProportion: Optional[float] = Field(
        None, description='Proportion of number of unique values in a column.'
    )
    nullCount: Optional[float] = Field(
        None, description='No.of null values in a column.'
    )
    nullProportion: Optional[float] = Field(
        None, description='No.of null value proportion in columns.'
    )
    min: Optional[str] = Field(None, description='Minimum value in a column.')
    max: Optional[str] = Field(None, description='Maximum value in a column.')
    mean: Optional[str] = Field(None, description='Avg value in a column.')
    median: Optional[str] = Field(None, description='Median value in a column.')
    stddev: Optional[float] = Field(None, description='Standard deviation of a column.')


class ModelType(Enum):
    DBT = 'DBT'


class TableJoins(BaseModel):
    class Config:
        extra = Extra.forbid

    startDate: Optional[basic.Date] = Field(
        None, description='Date can be only from today going back to last 29 days.'
    )
    dayCount: Optional[int] = 1
    columnJoins: Optional[List[ColumnJoins]] = None


class TableProfile(BaseModel):
    class Config:
        extra = Extra.forbid

    profileDate: Optional[basic.Date] = Field(
        None, description='Data one which profile is taken.'
    )
    columnCount: Optional[float] = Field(
        None, description='No.of columns in the table.'
    )
    rowCount: Optional[float] = Field(None, description='No.of rows in the table.')
    columnProfile: Optional[List[ColumnProfile]] = Field(
        None, description='List of local column profiles of the table.'
    )


class Column(BaseModel):
    class Config:
        extra = Extra.forbid

    name: ColumnName
    dataType: DataType = Field(
        ..., description='Data type of the column (int, date etc.).'
    )
    arrayDataType: Optional[DataType] = Field(
        None,
        description='Data type used array in dataType. For example, `array<int>` has dataType as `array` and arrayDataType as `int`.',
    )
    dataLength: Optional[int] = Field(
        None,
        description='Length of `char`, `varchar`, `binary`, `varbinary` `dataTypes`, else null. For example, `varchar(20)` has dataType as `varchar` and dataLength as `20`.',
    )
    dataTypeDisplay: Optional[str] = Field(
        None,
        description='Display name used for dataType. This is useful for complex types, such as `array<int>, map<int,string>, struct<>, and union types.',
    )
    description: Optional[str] = Field(None, description='Description of the column.')
    fullyQualifiedName: Optional[FullyQualifiedColumnName] = None
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags associated with the column.'
    )
    constraint: Optional[Constraint] = Field(
        None, description='Column level constraint.'
    )
    ordinalPosition: Optional[int] = Field(
        None, description='Ordinal position of the column.'
    )
    jsonSchema: Optional[str] = Field(
        None, description='Json schema only if the dataType is JSON else null.'
    )
    children: Optional[List[Column]] = Field(
        None,
        description='Child columns if dataType or arrayDataType is `map`, `struct`, or `union` else `null`.',
    )


class SqlQuery(BaseModel):
    query: Optional[str] = Field(
        None, description='SQL Query text that matches the table name.'
    )
    duration: Optional[float] = Field(
        None, description='How long did the query took to run in seconds.'
    )
    user: Optional[entityReference.EntityReference] = Field(
        None, description='User who ran this query.'
    )
    vote: Optional[float] = Field(
        1, description='Users can vote up to rank the popular queries.'
    )
    checksum: Optional[str] = Field(
        None, description='Checksum to avoid registering duplicate queries.'
    )
    queryDate: Optional[basic.Date] = Field(
        None, description='Date on which the query ran.'
    )


class DataModel(BaseModel):
    modelType: ModelType
    description: Optional[str] = Field(
        None, description='Description of the Table from the model.'
    )
    path: Optional[str] = Field(None, description='Path to sql definition file.')
    rawSql: Optional[basic.SqlQuery] = Field(
        None,
        description='This corresponds to rws SQL from `<model_name>.sql` in DBT. This might be null when SQL query need not be compiled as done in DBT.',
    )
    sql: basic.SqlQuery = Field(
        ...,
        description='This corresponds to compile SQL from `<model_name>.sql` in DBT. In cases where compilation is not necessary, this corresponds to SQL that created the table.',
    )
    upstream: Optional[List[str]] = Field(
        None,
        description='Fully qualified name of Models/tables used for in `sql` for creating this table.',
    )
    columns: Optional[List[Column]] = Field(
        None,
        description='Columns from the schema defined during modeling. In case of DBT, the metadata here comes from `schema.yaml`.',
    )
    generatedAt: Optional[basic.DateTime] = None


class Table(BaseModel):
    class Config:
        extra = Extra.forbid

    id: basic.Uuid = Field(..., description='Unique identifier of this table instance.')
    name: TableName = Field(
        ..., description='Name of a table. Expected to be unique within a database.'
    )
    displayName: Optional[str] = Field(
        None,
        description='Display Name that identifies this table. It could be title or label from the source services.',
    )
    fullyQualifiedName: Optional[str] = Field(
        None,
        description='Fully qualified name of a table in the form `serviceName.databaseName.tableName`.',
    )
    description: Optional[str] = Field(None, description='Description of a table.')
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    href: Optional[basic.Href] = Field(None, description='Link to this table resource.')
    tableType: Optional[TableType] = None
    columns: List[Column] = Field(..., description='Columns in this table.')
    tableConstraints: Optional[List[TableConstraint]] = Field(
        None, description='Table constraints.'
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this table.'
    )
    database: Optional[entityReference.EntityReference] = Field(
        None, description='Reference to Database that contains this table.'
    )
    service: Optional[entityReference.EntityReference] = Field(
        None, description='Link to Database service this table is hosted in.'
    )
    serviceType: Optional[databaseService.DatabaseServiceType] = Field(
        None, description='Service type this table is hosted in.'
    )
    location: Optional[entityReference.EntityReference] = Field(
        None, description='Reference to the Location that contains this table.'
    )
    viewDefinition: Optional[basic.SqlQuery] = Field(
        None, description='View Definition in SQL. Applies to TableType.View only.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this table.'
    )
    usageSummary: Optional[usageDetails.TypeUsedToReturnUsageDetailsOfAnEntity] = Field(
        None, description='Latest usage information for this table.'
    )
    followers: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Followers of this table.'
    )
    joins: Optional[TableJoins] = Field(
        None,
        description='Details of other tables this table is frequently joined with.',
    )
    sampleData: Optional[TableData] = Field(
        None, description='Sample data for a table.'
    )
    tableProfile: Optional[List[TableProfile]] = Field(
        None, description='Data profile for a table.'
    )
    tableQueries: Optional[List[SqlQuery]] = Field(
        None, description='List of queries that ran against a table.'
    )
    dataModel: Optional[DataModel] = Field(
        None,
        description='This captures information about how the table is modeled. Currently only DBT model is supported.',
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that lead to this version of the entity.'
    )


Column.update_forward_refs()



# generated by datamodel-codegen:
#   filename:  schema/entity/services/dashboardService.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import Optional

from pydantic import AnyUrl, BaseModel, Extra, Field, constr

from ...type import basic, entityHistory, schedule


class DashboardServiceType(Enum):
    Superset = 'Superset'
    Looker = 'Looker'
    Tableau = 'Tableau'
    Redash = 'Redash'
    Metabase = 'Metabase'


class DashboardService(BaseModel):
    class Config:
        extra = Extra.forbid

    id: basic.Uuid = Field(
        ..., description='Unique identifier of this dashboard service instance.'
    )
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this dashboard service.'
    )
    displayName: Optional[str] = Field(
        None, description='Display Name that identifies this dashboard service.'
    )
    serviceType: DashboardServiceType = Field(
        ..., description='Type of dashboard service such as Looker or Superset...'
    )
    description: Optional[str] = Field(
        None, description='Description of a dashboard service instance.'
    )
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    dashboardUrl: AnyUrl = Field(
        ...,
        description='Dashboard Service URL. This will be used to make REST API calls to Dashboard Service.',
    )
    username: Optional[str] = Field(
        None, description='Username to log-into Dashboard Service.'
    )
    password: Optional[str] = Field(
        None, description='Password to log-into Dashboard Service.'
    )
    ingestionSchedule: Optional[schedule.Schedule] = Field(
        None, description='Schedule for running metadata ingestion jobs.'
    )
    href: Optional[basic.Href] = Field(
        None,
        description='Link to the resource corresponding to this dashboard service.',
    )
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that lead to this version of the entity.'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This import verifies that the dependencies are available.
import logging
from typing import Any, Dict, Iterable, Iterator, Union

from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.table_queries import TableQuery
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.redshift import RedshiftConfig
from metadata.ingestion.source.sql_alchemy_helper import (
    SQLAlchemyHelper,
    SQLSourceStatus,
)
from metadata.utils.helpers import get_start_and_end

logger = logging.getLogger(__name__)


class RedshiftUsageSource(Source[TableQuery]):
    # SELECT statement from mysql information_schema to extract table and column metadata
    SQL_STATEMENT = """
        SELECT DISTINCT ss.userid,
            ss.query,
            sui.usename,
            ss.tbl,
            sq.querytxt,
            sti.database,
            sti.schema,
            sti.table,
            sq.starttime,
            sq.endtime,
            sq.aborted
        FROM stl_scan ss
            JOIN svv_table_info sti ON ss.tbl = sti.table_id
            JOIN stl_query sq ON ss.query = sq.query
            JOIN svl_user_info sui ON sq.userid = sui.usesysid
        WHERE ss.starttime >= '{start_time}'
            AND ss.starttime < '{end_time}'
            AND sq.aborted = 0
        ORDER BY ss.endtime DESC;
    """
    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = "where_clause"
    CLUSTER_SOURCE = "cluster_source"
    CLUSTER_KEY = "cluster_key"
    USE_CATALOG_AS_CLUSTER_NAME = "use_catalog_as_cluster_name"
    DATABASE_KEY = "database_key"
    SERVICE_TYPE = "Redshift"
    DEFAULT_CLUSTER_SOURCE = "CURRENT_DATABASE()"

    def __init__(self, config, metadata_config, ctx):
        super().__init__(ctx)
        start, end = get_start_and_end(config.duration)
        self.sql_stmt = RedshiftUsageSource.SQL_STATEMENT.format(
            start_time=start, end_time=end
        )
        self.analysis_date = start
        self.alchemy_helper = SQLAlchemyHelper(
            config, metadata_config, ctx, "Redshift", self.sql_stmt
        )
        self._extract_iter: Union[None, Iterator] = None
        self._database = "redshift"
        self.status = SQLSourceStatus()

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = RedshiftConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def _get_raw_extract_iter(self) -> Iterable[Dict[str, Any]]:
        """
        Provides iterator of result row from SQLAlchemy helper
        :return:
        """
        rows = self.alchemy_helper.execute_query()
        for row in rows:
            yield row

    def next_record(self) -> Iterable[TableQuery]:
        """
        Using itertools.groupby and raw level iterator, it groups to table and yields TableMetadata
        :return:
        """
        for row in self._get_raw_extract_iter():
            tq = TableQuery(
                query=row["query"],
                user_name=row["usename"],
                starttime=str(row["starttime"]),
                endtime=str(row["endtime"]),
                analysis_date=str(self.analysis_date),
                database=row["database"],
                aborted=row["aborted"],
                sql=row["querytxt"],
            )
            yield tq

    def close(self):
        self.alchemy_helper.close()

    def get_status(self) -> SourceStatus:
        return self.status



# generated by datamodel-codegen:
#   filename:  schema/api/tags/createTagCategory.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from pydantic import BaseModel, Field

from ...entity.tags import tagCategory


class CreateTagCategoryEntityRequest(BaseModel):
    name: tagCategory.TagName
    description: str = Field(..., description='Description of the tag category')
    categoryType: tagCategory.TagCategoryType



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import ssl
import time
from datetime import datetime
from typing import List, Optional

from dateutil import parser
from elasticsearch import Elasticsearch
from elasticsearch.connection import create_ssl_context

from metadata.config.common import ConfigModel
from metadata.generated.schema.entity.data.chart import Chart
from metadata.generated.schema.entity.data.dashboard import Dashboard
from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.pipeline import Pipeline, Task
from metadata.generated.schema.entity.data.table import Column, Table
from metadata.generated.schema.entity.data.topic import Topic
from metadata.generated.schema.entity.services.dashboardService import DashboardService
from metadata.generated.schema.entity.services.databaseService import DatabaseService
from metadata.generated.schema.entity.services.messagingService import MessagingService
from metadata.generated.schema.entity.services.pipelineService import PipelineService
from metadata.generated.schema.type import entityReference
from metadata.ingestion.api.common import Entity, WorkflowContext
from metadata.ingestion.api.sink import Sink, SinkStatus
from metadata.ingestion.models.table_metadata import (
    ChangeDescription,
    DashboardESDocument,
    PipelineESDocument,
    TableESDocument,
    TopicESDocument,
)
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.sink.elasticsearch_constants import (
    DASHBOARD_ELASTICSEARCH_INDEX_MAPPING,
    PIPELINE_ELASTICSEARCH_INDEX_MAPPING,
    TABLE_ELASTICSEARCH_INDEX_MAPPING,
    TOPIC_ELASTICSEARCH_INDEX_MAPPING,
)

logger = logging.getLogger(__name__)


def epoch_ms(dt: datetime):
    return int(dt.timestamp() * 1000)


class ElasticSearchConfig(ConfigModel):
    es_host: str
    es_port: int = 9200
    es_username: Optional[str] = None
    es_password: Optional[str] = None
    index_tables: Optional[bool] = True
    index_topics: Optional[bool] = True
    index_dashboards: Optional[bool] = True
    index_pipelines: Optional[bool] = True
    index_dbt_models: Optional[bool] = True
    table_index_name: str = "table_search_index"
    topic_index_name: str = "topic_search_index"
    dashboard_index_name: str = "dashboard_search_index"
    pipeline_index_name: str = "pipeline_search_index"
    dbt_index_name: str = "dbt_model_search_index"
    scheme: str = "http"
    use_ssl: bool = False
    verify_certs: bool = False
    timeout: int = 30
    ca_certs: Optional[str] = None


class ElasticsearchSink(Sink[Entity]):
    """ """

    DEFAULT_ELASTICSEARCH_INDEX_MAPPING = TABLE_ELASTICSEARCH_INDEX_MAPPING

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = ElasticSearchConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(ctx, config, metadata_config)

    def __init__(
        self,
        ctx: WorkflowContext,
        config: ElasticSearchConfig,
        metadata_config: MetadataServerConfig,
    ) -> None:

        self.config = config
        self.metadata_config = metadata_config
        self.ctx = ctx
        self.status = SinkStatus()
        self.metadata = OpenMetadata(self.metadata_config)
        self.elasticsearch_doc_type = "_doc"
        http_auth = None
        if self.config.es_username:
            http_auth = (self.config.es_username, self.config.es_password)

        ssl_context = None
        if self.config.scheme == "https" and not self.config.verify_certs:
            ssl_context = create_ssl_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

        self.elasticsearch_client = Elasticsearch(
            [
                {"host": self.config.es_host, "port": self.config.es_port},
            ],
            http_auth=http_auth,
            scheme=self.config.scheme,
            use_ssl=self.config.use_ssl,
            verify_certs=self.config.verify_certs,
            ssl_context=ssl_context,
            ca_certs=self.config.ca_certs,
        )

        if self.config.index_tables:
            self._check_or_create_index(
                self.config.table_index_name, TABLE_ELASTICSEARCH_INDEX_MAPPING
            )
        if self.config.index_topics:
            self._check_or_create_index(
                self.config.topic_index_name, TOPIC_ELASTICSEARCH_INDEX_MAPPING
            )
        if self.config.index_dashboards:
            self._check_or_create_index(
                self.config.dashboard_index_name, DASHBOARD_ELASTICSEARCH_INDEX_MAPPING
            )
        if self.config.index_pipelines:
            self._check_or_create_index(
                self.config.pipeline_index_name, PIPELINE_ELASTICSEARCH_INDEX_MAPPING
            )

    def _check_or_create_index(self, index_name: str, es_mapping: str):
        """
        Retrieve all indices that currently have {elasticsearch_alias} alias
        :return: list of elasticsearch indices
        """
        if self.elasticsearch_client.indices.exists(index_name):
            mapping = self.elasticsearch_client.indices.get_mapping()
            if not mapping[index_name]["mappings"]:
                logger.debug(
                    f"There are no mappings for index {index_name}. Updating the mapping"
                )
                es_mapping_dict = json.loads(es_mapping)
                es_mapping_update_dict = {
                    "properties": es_mapping_dict["mappings"]["properties"]
                }
                self.elasticsearch_client.indices.put_mapping(
                    index=index_name,
                    body=json.dumps(es_mapping_update_dict),
                    request_timeout=self.config.timeout,
                )
        else:
            logger.warning(
                "Received index not found error from Elasticsearch. "
                + "The index doesn't exist for a newly created ES. It's OK on first run."
            )
            # create new index with mapping
            self.elasticsearch_client.indices.create(
                index=index_name, body=es_mapping, request_timeout=self.config.timeout
            )

    def write_record(self, record: Entity) -> None:
        if isinstance(record, Table):
            table_doc = self._create_table_es_doc(record)
            self.elasticsearch_client.index(
                index=self.config.table_index_name,
                id=str(table_doc.table_id),
                body=table_doc.json(),
                request_timeout=self.config.timeout,
            )
        if isinstance(record, Topic):
            topic_doc = self._create_topic_es_doc(record)
            self.elasticsearch_client.index(
                index=self.config.topic_index_name,
                id=str(topic_doc.topic_id),
                body=topic_doc.json(),
                request_timeout=self.config.timeout,
            )
        if isinstance(record, Dashboard):
            dashboard_doc = self._create_dashboard_es_doc(record)
            self.elasticsearch_client.index(
                index=self.config.dashboard_index_name,
                id=str(dashboard_doc.dashboard_id),
                body=dashboard_doc.json(),
                request_timeout=self.config.timeout,
            )
        if isinstance(record, Pipeline):
            pipeline_doc = self._create_pipeline_es_doc(record)
            self.elasticsearch_client.index(
                index=self.config.pipeline_index_name,
                id=str(pipeline_doc.pipeline_id),
                body=pipeline_doc.json(),
                request_timeout=self.config.timeout,
            )

        if hasattr(record.name, "__root__"):
            self.status.records_written(record.name.__root__)
        else:
            self.status.records_written(record.name)

    def _create_table_es_doc(self, table: Table):
        fqdn = table.fullyQualifiedName
        database = table.database.name
        table_name = table.name
        suggest = [
            {"input": [fqdn], "weight": 5},
            {"input": [table_name], "weight": 10},
        ]
        column_names = []
        column_descriptions = []
        tags = set()

        timestamp = epoch_ms(table.updatedAt.__root__)
        tier = None
        for table_tag in table.tags:
            if "Tier" in table_tag.tagFQN:
                tier = table_tag.tagFQN
            else:
                tags.add(table_tag.tagFQN)
        self._parse_columns(
            table.columns, None, column_names, column_descriptions, tags
        )

        database_entity = self.metadata.get_by_id(
            entity=Database, entity_id=str(table.database.id.__root__)
        )
        service_entity = self.metadata.get_by_id(
            entity=DatabaseService, entity_id=str(database_entity.service.id.__root__)
        )
        table_owner = str(table.owner.id.__root__) if table.owner is not None else ""
        table_followers = []
        if table.followers:
            for follower in table.followers.__root__:
                table_followers.append(str(follower.id.__root__))
        table_type = None
        if hasattr(table.tableType, "name"):
            table_type = table.tableType.name

        change_descriptions = self._get_change_descriptions(Table, table.id.__root__)
        table_doc = TableESDocument(
            table_id=str(table.id.__root__),
            database=str(database_entity.name.__root__),
            service=service_entity.name,
            service_type=service_entity.serviceType.name,
            service_category="databaseService",
            name=table.name.__root__,
            suggest=suggest,
            description=table.description,
            table_type=table_type,
            last_updated_timestamp=timestamp,
            column_names=column_names,
            column_descriptions=column_descriptions,
            monthly_stats=table.usageSummary.monthlyStats.count,
            monthly_percentile_rank=table.usageSummary.monthlyStats.percentileRank,
            weekly_stats=table.usageSummary.weeklyStats.count,
            weekly_percentile_rank=table.usageSummary.weeklyStats.percentileRank,
            daily_stats=table.usageSummary.dailyStats.count,
            daily_percentile_rank=table.usageSummary.dailyStats.percentileRank,
            tier=tier,
            tags=list(tags),
            fqdn=fqdn,
            owner=table_owner,
            followers=table_followers,
            change_descriptions=change_descriptions,
        )
        return table_doc

    def _create_topic_es_doc(self, topic: Topic):
        fqdn = topic.fullyQualifiedName
        topic_name = topic.name
        suggest = [
            {"input": [fqdn], "weight": 5},
            {"input": [topic_name], "weight": 10},
        ]
        tags = set()
        timestamp = epoch_ms(topic.updatedAt.__root__)
        service_entity = self.metadata.get_by_id(
            entity=MessagingService, entity_id=str(topic.service.id.__root__)
        )
        topic_owner = str(topic.owner.id.__root__) if topic.owner is not None else ""
        topic_followers = []
        if topic.followers:
            for follower in topic.followers.__root__:
                topic_followers.append(str(follower.id.__root__))
        tier = None
        for topic_tag in topic.tags:
            if "Tier" in topic_tag.tagFQN:
                tier = topic_tag.tagFQN
            else:
                tags.add(topic_tag.tagFQN)
        change_descriptions = self._get_change_descriptions(Topic, topic.id.__root__)
        topic_doc = TopicESDocument(
            topic_id=str(topic.id.__root__),
            service=service_entity.name,
            service_type=service_entity.serviceType.name,
            service_category="messagingService",
            name=topic.name.__root__,
            suggest=suggest,
            description=topic.description,
            last_updated_timestamp=timestamp,
            tier=tier,
            tags=list(tags),
            fqdn=fqdn,
            owner=topic_owner,
            followers=topic_followers,
            change_descriptions=change_descriptions,
        )
        return topic_doc

    def _create_dashboard_es_doc(self, dashboard: Dashboard):
        fqdn = dashboard.fullyQualifiedName
        dashboard_name = dashboard.name
        suggest = [{"input": [dashboard.displayName], "weight": 10}]
        tags = set()
        timestamp = epoch_ms(dashboard.updatedAt.__root__)
        service_entity = self.metadata.get_by_id(
            entity=DashboardService, entity_id=str(dashboard.service.id.__root__)
        )
        dashboard_owner = (
            str(dashboard.owner.id.__root__) if dashboard.owner is not None else ""
        )
        dashboard_followers = []
        if dashboard.followers:
            for follower in dashboard.followers.__root__:
                dashboard_followers.append(str(follower.id.__root__))
        tier = None
        for dashboard_tag in dashboard.tags:
            if "Tier" in dashboard_tag.tagFQN:
                tier = dashboard_tag.tagFQN
            else:
                tags.add(dashboard_tag.tagFQN)
        charts: List[Chart] = self._get_charts(dashboard.charts)
        chart_names = []
        chart_descriptions = []
        for chart in charts:
            chart_names.append(chart.displayName)
            if chart.description is not None:
                chart_descriptions.append(chart.description)
            if len(chart.tags) > 0:
                for col_tag in chart.tags:
                    tags.add(col_tag.tagFQN)
        change_descriptions = self._get_change_descriptions(
            Dashboard, dashboard.id.__root__
        )
        dashboard_doc = DashboardESDocument(
            dashboard_id=str(dashboard.id.__root__),
            service=service_entity.name,
            service_type=service_entity.serviceType.name,
            service_category="dashboardService",
            name=dashboard.displayName,
            chart_names=chart_names,
            chart_descriptions=chart_descriptions,
            suggest=suggest,
            description=dashboard.description,
            last_updated_timestamp=timestamp,
            tier=tier,
            tags=list(tags),
            fqdn=fqdn,
            owner=dashboard_owner,
            followers=dashboard_followers,
            monthly_stats=dashboard.usageSummary.monthlyStats.count,
            monthly_percentile_rank=dashboard.usageSummary.monthlyStats.percentileRank,
            weekly_stats=dashboard.usageSummary.weeklyStats.count,
            weekly_percentile_rank=dashboard.usageSummary.weeklyStats.percentileRank,
            daily_stats=dashboard.usageSummary.dailyStats.count,
            daily_percentile_rank=dashboard.usageSummary.dailyStats.percentileRank,
            change_descriptions=change_descriptions,
        )

        return dashboard_doc

    def _create_pipeline_es_doc(self, pipeline: Pipeline):
        fqdn = pipeline.fullyQualifiedName
        suggest = [{"input": [pipeline.displayName], "weight": 10}]
        tags = set()
        timestamp = epoch_ms(pipeline.updatedAt.__root__)
        service_entity = self.metadata.get_by_id(
            entity=PipelineService, entity_id=str(pipeline.service.id.__root__)
        )
        pipeline_owner = (
            str(pipeline.owner.id.__root__) if pipeline.owner is not None else ""
        )
        pipeline_followers = []
        if pipeline.followers:
            for follower in pipeline.followers.__root__:
                pipeline_followers.append(str(follower.id.__root__))
        tier = None
        for pipeline_tag in pipeline.tags:
            if "Tier" in pipeline_tag.tagFQN:
                tier = pipeline_tag.tagFQN
            else:
                tags.add(pipeline_tag.tagFQN)
        tasks: List[Task] = pipeline.tasks
        task_names = []
        task_descriptions = []
        for task in tasks:
            task_names.append(task.displayName)
            if task.description is not None:
                task_descriptions.append(task.description)
            if tags in task and len(task.tags) > 0:
                for col_tag in task.tags:
                    tags.add(col_tag.tagFQN)
        change_descriptions = self._get_change_descriptions(
            Pipeline, pipeline.id.__root__
        )
        pipeline_doc = PipelineESDocument(
            pipeline_id=str(pipeline.id.__root__),
            service=service_entity.name,
            service_type=service_entity.serviceType.name,
            service_category="pipelineService",
            name=pipeline.displayName,
            task_names=task_names,
            task_descriptions=task_descriptions,
            suggest=suggest,
            description=pipeline.description,
            last_updated_timestamp=timestamp,
            tier=tier,
            tags=list(tags),
            fqdn=fqdn,
            owner=pipeline_owner,
            followers=pipeline_followers,
            change_descriptions=change_descriptions,
        )

        return pipeline_doc

    def _get_charts(self, chart_refs: Optional[List[entityReference.EntityReference]]):
        charts = []
        if chart_refs:
            for chart_ref in chart_refs:
                chart = self.metadata.get_by_id(
                    entity=Chart, entity_id=str(chart_ref.id.__root__), fields=["tags"]
                )
                charts.append(chart)
        return charts

    def _parse_columns(
        self,
        columns: List[Column],
        parent_column,
        column_names,
        column_descriptions,
        tags,
    ):
        for column in columns:
            col_name = (
                parent_column + "." + column.name.__root__
                if parent_column is not None
                else column.name.__root__
            )
            column_names.append(col_name)
            if column.description is not None:
                column_descriptions.append(column.description)
            if len(column.tags) > 0:
                for col_tag in column.tags:
                    tags.add(col_tag.tagFQN)
            if column.children is not None:
                self._parse_columns(
                    column.children,
                    column.name.__root__,
                    column_names,
                    column_descriptions,
                    tags,
                )

    def _get_change_descriptions(self, entity_type, entity_id):
        try:
            entity_versions = self.metadata.list_versions(entity_id, entity_type)
            change_descriptions = []
            for version in entity_versions.versions:
                version_json = json.loads(version)
                updatedAt = parser.parse(version_json["updatedAt"])
                change_description = ChangeDescription(
                    updatedBy=version_json["updatedBy"], updatedAt=epoch_ms(updatedAt)
                )
                if "changeDescription" in version_json:
                    change_description.fieldsAdded = version_json["changeDescription"][
                        "fieldsAdded"
                    ]
                    change_description.fieldsDeleted = version_json[
                        "changeDescription"
                    ]["fieldsDeleted"]
                    change_description.fieldsUpdated = version_json[
                        "changeDescription"
                    ]["fieldsUpdated"]
                change_descriptions.append(change_description)
            return change_descriptions
        except Exception as err:
            logger.error(repr(err))

    def get_status(self):
        return self.status

    def close(self):
        self.elasticsearch_client.close()



# generated by datamodel-codegen:
#   filename:  schema/api/data/createDatabase.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field

from ...entity.data import database
from ...type import entityReference


class CreateDatabaseEntityRequest(BaseModel):
    name: database.DatabaseName = Field(
        ..., description='Name that identifies this database instance uniquely.'
    )
    description: Optional[str] = Field(
        None,
        description='Description of the database instance. What it has and how to use it.',
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this database'
    )
    service: entityReference.EntityReference = Field(
        ..., description='Link to the database service where this database is hosted in'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime, timedelta
from typing import List

from metadata.generated.schema.api.services.createDashboardService import (
    CreateDashboardServiceEntityRequest,
)
from metadata.generated.schema.api.services.createDatabaseService import (
    CreateDatabaseServiceEntityRequest,
)
from metadata.generated.schema.api.services.createMessagingService import (
    CreateMessagingServiceEntityRequest,
)
from metadata.generated.schema.api.services.createPipelineService import (
    CreatePipelineServiceEntityRequest,
)
from metadata.generated.schema.api.services.createStorageService import (
    CreateStorageServiceEntityRequest,
)
from metadata.generated.schema.entity.services.dashboardService import DashboardService
from metadata.generated.schema.entity.services.databaseService import DatabaseService
from metadata.generated.schema.entity.services.messagingService import MessagingService
from metadata.generated.schema.entity.services.pipelineService import PipelineService
from metadata.generated.schema.entity.services.storageService import StorageService
from metadata.ingestion.ometa.ometa_api import OpenMetadata


def get_start_and_end(duration):
    today = datetime.utcnow()
    start = (today + timedelta(0 - duration)).replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    end = (today + timedelta(3)).replace(hour=0, minute=0, second=0, microsecond=0)
    return start, end


def snake_to_camel(s):
    a = s.split("_")
    a[0] = a[0].capitalize()
    if len(a) > 1:
        a[1:] = [u.title() for u in a[1:]]
    return "".join(a)


def get_database_service_or_create(
    config, metadata_config, service_name=None
) -> DatabaseService:
    metadata = OpenMetadata(metadata_config)
    config.service_name = service_name if service_name else config.service_name
    service = metadata.get_by_name(entity=DatabaseService, fqdn=config.service_name)
    if service:
        return service
    else:
        service = {
            "jdbc": {
                "connectionUrl": f"jdbc://{config.host_port}",
                "driverClass": "jdbc",
            },
            "name": config.service_name,
            "description": "",
            "serviceType": config.get_service_type(),
        }
        created_service = metadata.create_or_update(
            CreateDatabaseServiceEntityRequest(**service)
        )
        return created_service


def get_messaging_service_or_create(
    service_name: str,
    message_service_type: str,
    schema_registry_url: str,
    brokers: List[str],
    metadata_config,
) -> MessagingService:
    metadata = OpenMetadata(metadata_config)
    service = metadata.get_by_name(entity=MessagingService, fqdn=service_name)
    if service is not None:
        return service
    else:
        created_service = metadata.create_or_update(
            CreateMessagingServiceEntityRequest(
                name=service_name,
                serviceType=message_service_type,
                brokers=brokers,
                schemaRegistry=schema_registry_url,
            )
        )
        return created_service


def get_dashboard_service_or_create(
    service_name: str,
    dashboard_service_type: str,
    username: str,
    password: str,
    dashboard_url: str,
    metadata_config,
) -> DashboardService:
    metadata = OpenMetadata(metadata_config)
    service = metadata.get_by_name(entity=DashboardService, fqdn=service_name)
    if service is not None:
        return service
    else:
        created_service = metadata.create_or_update(
            CreateDashboardServiceEntityRequest(
                name=service_name,
                serviceType=dashboard_service_type,
                username=username,
                password=password,
                dashboardUrl=dashboard_url,
            )
        )
        return created_service


def get_pipeline_service_or_create(service_json, metadata_config) -> PipelineService:
    metadata = OpenMetadata(metadata_config)
    service = metadata.get_by_name(entity=PipelineService, fqdn=service_json["name"])
    if service is not None:
        return service
    else:
        created_service = metadata.create_or_update(
            CreatePipelineServiceEntityRequest(**service_json)
        )
        return created_service


def get_storage_service_or_create(service_json, metadata_config) -> StorageService:
    metadata = OpenMetadata(metadata_config)
    service = metadata.get_by_name(entity=StorageService, fqdn=service_json["name"])
    if service is not None:
        return service
    else:
        created_service = metadata.create_or_update(
            CreateStorageServiceEntityRequest(**service_json)
        )
        return created_service


def get_database_service_or_create_v2(service_json, metadata_config) -> DatabaseService:
    metadata = OpenMetadata(metadata_config)
    service = metadata.get_by_name(entity=DatabaseService, fqdn=service_json["name"])
    if service is not None:
        return service
    else:
        created_service = metadata.create_or_update(
            CreateDatabaseServiceEntityRequest(**service_json)
        )
    return created_service


def convert_epoch_to_iso(seconds_since_epoch):
    dt = datetime.utcfromtimestamp(seconds_since_epoch)
    iso_format = dt.isoformat() + "Z"
    return iso_format



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable
from urllib.parse import quote_plus

from sqlalchemy.inspection import inspect

from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLConnectionConfig, SQLSource


class TrinoConfig(SQLConnectionConfig):
    host_port = "localhost:8080"
    scheme = "trino"
    service_type = "Trino"
    catalog: str
    database: str

    def get_connection_url(self):
        url = f"{self.scheme}://"
        if self.username is not None:
            url += f"{self.username}"
            if self.password is not None:
                url += f":{quote_plus(self.password.get_secret_value())}"
            url += "@"
        url += f"{self.host_port}"
        if self.catalog is not None:
            url += f"/{self.catalog}"
            if self.database is not None:
                url += f"/{self.database}"

        if self.options is not None:
            if self.database is None:
                url += "/"
            params = "&".join(
                f"{key}={quote_plus(value)}"
                for (key, value) in self.options.items()
                if value
            )
            url = f"{url}?{params}"
        return url


class TrinoSource(SQLSource):
    def __init__(self, config, metadata_config, ctx):
        super().__init__(config, metadata_config, ctx)

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = TrinoConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def next_record(self) -> Iterable[OMetaDatabaseAndTable]:
        inspector = inspect(self.engine)
        if self.config.include_tables:
            yield from self.fetch_tables(inspector, self.config.database)
        if self.config.include_views:
            yield from self.fetch_views(inspector, self.config.database)



# generated by datamodel-codegen:
#   filename:  schema/operations/workflows/ingestion.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Extra, Field, constr

from ...type import basic, entityHistory, entityReference, tagLabel


class IngestionType(Enum):
    bigquery = 'bigquery'
    bigquery_usage = 'bigquery-usage'
    redshift = 'redshift'
    redshift_usage = 'redshift-usage'
    snowflake = 'snowflake'
    snowflake_usage = 'snowflake-usage'
    hive = 'hive'
    mssql = 'mssql'
    mysql = 'mysql'
    postgres = 'postgres'
    presto = 'presto'
    trino = 'trino'
    vertica = 'vertica'


class ConnectorConfig(BaseModel):
    username: Optional[str] = Field(
        None, description='username to connect  to the data source.'
    )
    password: Optional[str] = Field(
        None, description='password to connect  to the data source.'
    )
    host: Optional[str] = Field(None, description='Host and port of the data source.')
    database: Optional[str] = Field(None, description='Database of the data source.')
    includeViews: Optional[bool] = Field(
        'true',
        description='optional configuration to turn off fetching metadata for views.',
    )
    enableDataProfiler: Optional[bool] = Field(
        'false',
        description='Run data profiler as part of ingestion to get table profile data.',
    )
    includeFilterPattern: Optional[List[str]] = Field(
        None,
        description='Regex to only fetch tables or databases that matches the pattern.',
    )
    excludeFilterPattern: Optional[List[str]] = Field(
        None, description='Regex exclude tables or databases that matches the pattern.'
    )


class IngestionStatus(BaseModel):
    state: Optional[str] = Field(
        None, description='Workflow status denotes if its failed or succeeded.'
    )
    startDate: Optional[str] = Field(
        None,
        description='startDate of the Ingestion Pipeline run for this particular execution.',
    )
    endDate: Optional[str] = Field(
        None,
        description='endDate of the Ingestion pipeline run for this particular execution.',
    )


class Ingestion(BaseModel):
    class Config:
        extra = Extra.forbid

    id: Optional[basic.Uuid] = Field(
        None, description='Unique identifier that identifies this Ingestion.'
    )
    name: constr(min_length=1, max_length=128) = Field(
        ..., description='Name that identifies this ingestion instance uniquely.'
    )
    displayName: Optional[str] = Field(
        None, description='Display Name that identifies this Ingestion.'
    )
    description: Optional[str] = Field(None, description='Description of the workflow.')
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this Ingestion.'
    )
    fullyQualifiedName: Optional[constr(min_length=1, max_length=256)] = Field(
        None, description='Name that uniquely identifies a Ingestion.'
    )
    ingestionType: Optional[IngestionType] = None
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags associated with the Ingestion.'
    )
    forceDeploy: Optional[bool] = Field(
        'false',
        description='Deploy the workflow by overwriting existing workflow with the same name.',
    )
    pauseWorkflow: Optional[bool] = Field(
        'false',
        description='pause the workflow from running once the deploy is finished successfully.',
    )
    concurrency: Optional[int] = Field(1, description='Concurrency of the Pipeline.')
    startDate: basic.Date = Field(..., description='Start date of the workflow.')
    endDate: Optional[basic.Date] = Field(None, description='End Date of the workflow.')
    nextExecutionDate: Optional[basic.Date] = Field(
        None,
        description='Next execution date from the underlying workflow platform once the ingestion scheduled.',
    )
    workflowTimezone: Optional[str] = Field(
        'UTC', description='Timezone in which workflow going to be scheduled.'
    )
    retries: Optional[int] = Field(1, description='Retry workflow in case of failure.')
    retryDelay: Optional[int] = Field(
        300, description='Delay between retries in seconds.'
    )
    workflowCatchup: Optional[bool] = Field(
        'false', description='Run past executions if the start date is in the past.'
    )
    scheduleInterval: Optional[str] = Field(
        None, description='Scheduler Interval for the Workflow in cron format.'
    )
    workflowTimeout: Optional[int] = Field(
        60, description='Timeout for the workflow in seconds.'
    )
    connectorConfig: ConnectorConfig
    ingestionStatuses: Optional[List[IngestionStatus]] = Field(
        None, description='List of executions and status for the Ingestion Pipeline.'
    )
    service: entityReference.EntityReference = Field(
        ...,
        description='Link to the database service where this database is hosted in.',
    )
    href: Optional[basic.Href] = Field(
        None, description='Link to this ingestion resource.'
    )
    version: Optional[entityHistory.EntityVersion] = Field(
        None, description='Metadata version of the entity.'
    )
    updatedAt: Optional[basic.DateTime] = Field(
        None,
        description='Last update time corresponding to the new version of the entity.',
    )
    updatedBy: Optional[str] = Field(None, description='User who made the update.')
    changeDescription: Optional[entityHistory.ChangeDescription] = Field(
        None, description='Change that led to this version of the entity.'
    )



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from datetime import datetime
from typing import Any, Iterable, Optional

from data_profiler.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from data_profiler.data_context import BaseDataContext
from data_profiler.data_context.types.base import (
    DataContextConfig,
    DatasourceConfig,
    InMemoryStoreBackendDefaults,
)

from metadata.generated.schema.entity.data.table import ColumnProfile, TableProfile
from metadata.ingestion.api.source import SourceStatus
from metadata.profiler.util import group_by

logger: logging.Logger = logging.getLogger(__name__)


class DataProfiler:
    data_context: BaseDataContext
    status: SourceStatus
    datasource_name: str = "om_data_source"

    def __init__(self, connection_str, status):
        self.status = status
        self.connection_str = connection_str
        data_context_config = DataContextConfig(
            datasources={
                self.datasource_name: DatasourceConfig(
                    class_name="SqlAlchemyDatasource",
                    credentials={
                        "url": self.connection_str,
                    },
                )
            },
            store_backend_defaults=InMemoryStoreBackendDefaults(),
            anonymous_usage_statistics={
                "enabled": False,
            },
        )

        self.data_context = BaseDataContext(project_config=data_context_config)

    def run_profiler(
        self,
        dataset_name: str,
        profile_date: str,
        schema: str = None,
        table: str = None,
        limit: int = None,
        offset: int = None,
        **kwargs: Any,
    ) -> TableProfile:
        try:
            profile_test_results = self._profile_data_asset(
                {
                    "schema": schema,
                    "table": table,
                    "limit": limit,
                    "offset": offset,
                    **kwargs,
                }
            )
            profile = self._parse_test_results_to_table_profile(
                profile_test_results,
                dataset_name=dataset_name,
                profile_date=profile_date,
            )
            return profile
        except Exception as err:
            logger.error(err)
            pass

    def _profile_data_asset(
        self, batch_kwargs: dict
    ) -> ExpectationSuiteValidationResult:

        profile_results = self.data_context.profile_data_asset(
            self.datasource_name,
            batch_kwargs={
                "datasource": self.datasource_name,
                **batch_kwargs,
            },
        )
        assert profile_results["success"]

        assert len(profile_results["results"]) == 1
        test_suite, test_results = profile_results["results"][0]
        return test_results

    @staticmethod
    def _get_column_from_result(result: ExpectationValidationResult) -> Optional[str]:
        return result.expectation_config.kwargs.get("column")

    def _parse_test_results_to_table_profile(
        self,
        profile_test_results: ExpectationSuiteValidationResult,
        dataset_name: str,
        profile_date: str,
    ) -> TableProfile:
        profile = None
        column_profiles = []
        for col, col_test_result in group_by(
            profile_test_results.results, key=self._get_column_from_result
        ):
            if col is None:
                profile = self._parse_table_test_results(
                    col_test_result,
                    dataset_name=dataset_name,
                    profile_date=profile_date,
                )
            else:
                column_profile = self._parse_column_test_results(
                    col, col_test_result, dataset_name=dataset_name
                )
                column_profiles.append(column_profile)

        if profile is not None:
            profile.columnProfile = column_profiles
        return profile

    def _parse_table_test_results(
        self,
        table_test_results: Iterable[ExpectationValidationResult],
        dataset_name: str,
        profile_date: str,
    ) -> TableProfile:
        profile = TableProfile(profileDate=profile_date)
        for table_result in table_test_results:
            expectation: str = table_result.expectation_config.expectation_type
            result: dict = table_result.result
            if expectation == "expect_table_row_count_to_be_between":
                profile.rowCount = result["observed_value"]
            elif expectation == "expect_table_columns_to_match_ordered_list":
                profile.columnCount = len(result["observed_value"])
            else:
                self.status.warning(
                    f"profile of {dataset_name}", f"unknown table mapper {expectation}"
                )
        return profile

    def _parse_column_test_results(
        self,
        column: str,
        col_test_results: Iterable[ExpectationValidationResult],
        dataset_name: str,
    ) -> ColumnProfile:
        column_profile = ColumnProfile(name=column)
        for col_result in col_test_results:
            expectation: str = col_result.expectation_config.expectation_type
            result: dict = col_result.result
            if not result:
                self.status.warning(
                    f"profile of {dataset_name}",
                    f"{expectation} did not yield any results",
                )
                continue

            if expectation == "expect_column_unique_value_count_to_be_between":
                column_profile.uniqueCount = result["observed_value"]
            elif (
                expectation == "expect_column_proportion_of_unique_values_to_be_between"
            ):
                column_profile.uniqueProportion = result["observed_value"]
            elif expectation == "expect_column_values_to_not_be_null":
                column_profile.nullCount = result["unexpected_count"]
                if (
                    "unexpected_percent" in result
                    and result["unexpected_percent"] is not None
                ):
                    column_profile.nullProportion = result["unexpected_percent"] / 100
            elif expectation == "expect_column_values_to_not_match_regex":
                pass
            elif expectation == "expect_column_mean_to_be_between":
                column_profile.mean = str(result["observed_value"])
            elif expectation == "expect_column_min_to_be_between":
                column_profile.min = str(result["observed_value"])
            elif expectation == "expect_column_max_to_be_between":
                column_profile.max = str(result["observed_value"])
            elif expectation == "expect_column_median_to_be_between":
                column_profile.median = str(result["observed_value"])
            elif expectation == "expect_column_stdev_to_be_between":
                column_profile.stddev = str(result["observed_value"])
            elif expectation == "expect_column_quantile_values_to_be_between":
                pass
            elif expectation == "expect_column_values_to_be_in_set":
                # column_profile.sample_values = [
                #   str(v) for v in result["partial_unexpected_list"]
                # ]
                pass
            elif expectation == "expect_column_kl_divergence_to_be_less_than":
                pass
            elif expectation == "expect_column_distinct_values_to_be_in_set":
                pass
            elif expectation == "expect_column_values_to_be_in_type_list":
                pass
            elif expectation == "expect_column_values_to_be_unique":
                pass
            else:
                self.status.warning(
                    f"profile of {dataset_name}",
                    f"warning: unknown column mapper {expectation} in col {column}",
                )
        return column_profile



# generated by datamodel-codegen:
#   filename:  schema/type/collectionDescriptor.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import AnyUrl, BaseModel, Field

from . import profile


class CollectionInfo(BaseModel):
    name: Optional[str] = Field(
        None, description='Unique name that identifies a collection.'
    )
    documentation: Optional[str] = Field(None, description='Description of collection.')
    href: Optional[AnyUrl] = Field(
        None,
        description='URL of the API endpoint where given collections are available.',
    )
    images: Optional[profile.ImageList] = None


class SchemaForCollectionDescriptor(BaseModel):
    collection: Optional[CollectionInfo] = None



# generated by datamodel-codegen:
#   filename:  schema/type/dailyCount.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from pydantic import BaseModel, Extra, Field, conint

from . import basic


class DailyCountOfSomeMeasurement(BaseModel):
    class Config:
        extra = Extra.forbid

    count: conint(ge=0) = Field(
        ..., description='Daily count of a measurement on the given date.'
    )
    date: basic.Date



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
from datetime import datetime

from metadata.config.common import ConfigModel
from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.table import ColumnJoins, Table, TableJoins
from metadata.ingestion.api.bulk_sink import BulkSink, BulkSinkStatus
from metadata.ingestion.api.common import WorkflowContext
from metadata.ingestion.models.table_queries import (
    ColumnJoinedWith,
    TableColumn,
    TableUsageCount,
    TableUsageRequest,
)
from metadata.ingestion.ometa.client import APIError
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig

logger = logging.getLogger(__name__)


class MetadataUsageSinkConfig(ConfigModel):
    filename: str
    service_name: str


class MetadataUsageBulkSink(BulkSink):
    config: MetadataUsageSinkConfig

    def __init__(
        self,
        ctx: WorkflowContext,
        config: MetadataUsageSinkConfig,
        metadata_config: MetadataServerConfig,
    ):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.wrote_something = False
        self.file_handler = open(self.config.filename, "r")
        self.metadata = OpenMetadata(self.metadata_config)
        self.status = BulkSinkStatus()
        self.table_join_dict = {}
        self.today = datetime.today().strftime("%Y-%m-%d")

    @classmethod
    def create(
        cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext
    ):
        config = MetadataUsageSinkConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(ctx, config, metadata_config)

    def handle_work_unit_start(self, wu):
        pass

    def handle_work_unit_end(self, wu):
        pass

    def write_records(self) -> None:
        usage_records = [json.loads(l) for l in self.file_handler.readlines()]
        for record in usage_records:
            table_usage = TableUsageCount(**json.loads(record))
            if "." in table_usage.table:
                table_usage.table = table_usage.table.split(".")[1]
            table_entity = self.__get_table_entity(
                table_usage.database, table_usage.table
            )
            if table_entity is not None:
                table_usage_request = TableUsageRequest(
                    date=table_usage.date, count=table_usage.count
                )
                try:
                    self.metadata.publish_table_usage(table_entity, table_usage_request)
                except Exception as err:
                    self.status.failures.append(table_usage_request)
                    logger.error(
                        "Failed to update usage for {} {}".format(
                            table_usage.table, err
                        )
                    )
                table_join_request = self.__get_table_joins(table_usage)
                logger.debug("table join request {}".format(table_join_request))
                try:
                    if (
                        table_join_request is not None
                        and len(table_join_request.columnJoins) > 0
                    ):
                        self.metadata.publish_frequently_joined_with(
                            table_entity, table_join_request
                        )
                except APIError as err:
                    self.status.failures.append(table_join_request)
                    logger.error(
                        "Failed to update query join for {}, {}".format(
                            table_usage.table, err
                        )
                    )

            else:
                logger.warning(
                    "Table does not exist, skipping usage publish {}, {}".format(
                        table_usage.table, table_usage.database
                    )
                )
        try:
            self.metadata.compute_percentile(Table, self.today)
            self.metadata.compute_percentile(Database, self.today)
        except APIError:
            logger.error("Failed to publish compute.percentile")

    def __get_table_joins(self, table_usage):
        table_joins: TableJoins = TableJoins(columnJoins=[], startDate=table_usage.date)
        column_joins_dict = {}
        for column_join in table_usage.joins:
            joined_with = {}
            if column_join.table_column is None or len(column_join.joined_with) == 0:
                continue
            if column_join.table_column.column in column_joins_dict.keys():
                joined_with = column_joins_dict[column_join.table_column.column]
            else:
                column_joins_dict[column_join.table_column.column] = {}
            main_column_fqdn = self.__get_column_fqdn(
                table_usage.database, column_join.table_column
            )
            for column in column_join.joined_with:
                joined_column_fqdn = self.__get_column_fqdn(
                    table_usage.database, column
                )
                if joined_column_fqdn in joined_with.keys():
                    column_joined_with = joined_with[joined_column_fqdn]
                    column_joined_with.joinCount += 1
                    joined_with[joined_column_fqdn] = column_joined_with
                elif joined_column_fqdn is not None:
                    joined_with[joined_column_fqdn] = ColumnJoinedWith(
                        fullyQualifiedName=joined_column_fqdn, joinCount=1
                    )
                else:
                    logger.info(
                        f"Skipping join columns for {column} {joined_column_fqdn}"
                    )
            column_joins_dict[column_join.table_column.column] = joined_with

        for key, value in column_joins_dict.items():
            table_joins.columnJoins.append(
                ColumnJoins(columnName=key, joinedWith=list(value.values()))
            )
        return table_joins

    def __get_column_fqdn(self, database: str, table_column: TableColumn):
        table_entity = self.__get_table_entity(database, table_column.table)
        if table_entity is None:
            return None
        for tbl_column in table_entity.columns:
            if table_column.column.lower() == tbl_column.name.__root__.lower():
                return tbl_column.fullyQualifiedName.__root__

    def __get_table_entity(self, database_name: str, table_name: str) -> Table:
        table_fqn = f"{self.config.service_name}.{database_name}.{table_name}"
        table_entity = self.metadata.get_by_name(Table, fqdn=table_fqn)
        return table_entity

    def get_status(self):
        return self.status

    def close(self):
        self.file_handler.close()
        self.metadata.close()



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import traceback
import uuid
from typing import Iterable, Optional

from boto3 import Session

from metadata.generated.schema.entity.data.database import Database
from metadata.generated.schema.entity.data.location import Location, LocationType
from metadata.generated.schema.entity.data.pipeline import Pipeline, Task
from metadata.generated.schema.entity.data.table import Column, Table
from metadata.generated.schema.entity.services.databaseService import (
    DatabaseServiceType,
)
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import ConfigModel, Entity, IncludeFilterPattern
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from metadata.ingestion.ometa.ometa_api import OpenMetadata
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.ingestion.source.sql_source import SQLSourceStatus
from metadata.utils.column_helpers import check_column_complex_type
from metadata.utils.helpers import (
    get_database_service_or_create,
    get_pipeline_service_or_create,
    get_storage_service_or_create,
)

logger: logging.Logger = logging.getLogger(__name__)


class GlueSourceConfig(ConfigModel):
    service_type = "Glue"
    aws_access_key_id: Optional[str]
    aws_secret_access_key: Optional[str]
    aws_session_token: Optional[str]
    endpoint_url: Optional[str]
    region_name: str
    service_name: str
    storage_service_name: str = "S3"
    pipeline_service_name: str
    # Glue doesn't have an host_port but the service definition requires it
    host_port: str = ""
    filter_pattern: IncludeFilterPattern = IncludeFilterPattern.allow_all()

    def get_service_type(self) -> DatabaseServiceType:
        return DatabaseServiceType[self.service_type]

    @property
    def glue(self):
        return self.get_glue_client()

    def get_glue_client(self):
        if (
            self.aws_access_key_id
            and self.aws_secret_access_key
            and self.aws_session_token
        ):
            session = Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                region_name=self.region_name,
            )
        elif self.aws_access_key_id and self.aws_secret_access_key:
            session = Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name,
            )
        else:
            session = Session(region_name=self.region_name)
        if self.endpoint_url is not None:
            return session.client(service_name="glue", endpoint_url=self.endpoint_url)
        else:
            return session.client(service_name="glue")


class GlueSource(Source[Entity]):
    def __init__(
        self, config: GlueSourceConfig, metadata_config: MetadataServerConfig, ctx
    ):
        super().__init__(ctx)
        self.status = SQLSourceStatus()
        self.config = config
        self.metadata_config = metadata_config
        self.metadata = OpenMetadata(metadata_config)
        self.service = get_database_service_or_create(
            config, metadata_config, self.config.service_name
        )
        self.storage_service = get_storage_service_or_create(
            {"name": self.config.storage_service_name, "serviceType": "S3"},
            metadata_config,
        )
        self.task_id_mapping = {}
        self.pipeline_service = get_pipeline_service_or_create(
            {
                "name": self.config.pipeline_service_name,
                "serviceType": "Glue",
                "pipelineUrl": self.config.endpoint_url
                if self.config.endpoint_url is not None
                else f"https://glue.{ self.config.region_name }.amazonaws.com",
            },
            metadata_config,
        )
        self.glue = config.glue
        self.database_name = None
        self.next_db_token = None

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = GlueSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def assign_next_token_db(self, glue_db_resp):
        if "NextToken" in glue_db_resp:
            self.next_db_token = glue_db_resp["NextToken"]
        else:
            self.next_db_token = "break"

    def next_record(self) -> Iterable[Entity]:
        while True:
            if self.next_db_token == "break":
                break
            elif self.next_db_token:
                glue_db_resp = self.glue.get_databases(
                    NextToken=self.next_db_token, ResourceShareType="ALL"
                )
                self.assign_next_token_db(glue_db_resp)
            else:
                glue_db_resp = self.glue.get_databases(ResourceShareType="ALL")
                self.assign_next_token_db(glue_db_resp)
            for db in glue_db_resp["DatabaseList"]:
                self.database_name = db["Name"]
                yield from self.ingest_tables()
        yield from self.ingest_pipelines()

    def get_columns(self, columnData):
        row_order = 0
        for column in columnData["Columns"]:
            if column["Type"].lower().startswith("union"):
                column["Type"] = column["Type"].replace(" ", "")
            (
                col_type,
                data_type_display,
                arr_data_type,
                children,
            ) = check_column_complex_type(
                self.status, self.dataset_name, column["Type"].lower(), column["Name"]
            )
            yield Column(
                name=column["Name"].replace(".", "_DOT_")[:128],
                description="",
                dataType=col_type,
                dataTypeDisplay="{}({})".format(col_type, 1)
                if data_type_display is None
                else f"{data_type_display}",
                ordinalPosition=row_order,
                children=children,
                arrayDataType=arr_data_type,
                dataLength=1,
            )
            row_order += 1

    def ingest_tables(self, next_tables_token=None) -> Iterable[OMetaDatabaseAndTable]:
        try:
            if next_tables_token is not None:
                glue_resp = self.glue.get_tables(
                    DatabaseName=self.database_name, NextToken=next_tables_token
                )
            else:
                glue_resp = self.glue.get_tables(DatabaseName=self.database_name)
            for table in glue_resp["TableList"]:
                if not self.config.filter_pattern.included(table["Name"]):
                    self.status.filter(
                        "{}.{}".format(self.config.get_service_name(), table["Name"]),
                        "Table pattern not allowed",
                    )
                    continue
                database_entity = Database(
                    name=table["DatabaseName"],
                    service=EntityReference(id=self.service.id, type="databaseService"),
                )
                fqn = f"{self.config.service_name}.{self.database_name}.{table['Name']}"
                self.dataset_name = fqn
                table_columns = self.get_columns(table["StorageDescriptor"])
                location_entity = Location(
                    name=table["StorageDescriptor"]["Location"],
                    locationType=LocationType.Table,
                    service=EntityReference(
                        id=self.storage_service.id, type="storageService"
                    ),
                )
                table_entity = Table(
                    id=uuid.uuid4(),
                    name=table["Name"][:128],
                    description=table["Description"]
                    if hasattr(table, "Description")
                    else "",
                    fullyQualifiedName=fqn,
                    columns=table_columns,
                )
                table_and_db = OMetaDatabaseAndTable(
                    table=table_entity,
                    database=database_entity,
                    location=location_entity,
                )
                yield table_and_db
            if "NextToken" in glue_resp:
                yield from self.ingest_tables(glue_resp["NextToken"])
        except Exception as err:
            logger.error(traceback.format_exc())
            logger.error(traceback.print_exc())
            logger.error(err)

    def get_downstream_tasks(self, task_unique_id, tasks):
        downstreamTasks = []
        for edges in tasks["Edges"]:
            if edges["SourceId"] == task_unique_id:
                if edges["DestinationId"] in self.task_id_mapping.values():
                    downstreamTasks.append(
                        list(self.task_id_mapping.keys())[
                            list(self.task_id_mapping.values()).index(
                                edges["DestinationId"]
                            )
                        ][:128]
                    )
        return downstreamTasks

    def get_tasks(self, tasks):
        taskList = []
        for task in tasks["Graph"]["Nodes"]:
            task_name = task["Name"][:128]
            self.task_id_mapping[task_name] = task["UniqueId"]
        for task in tasks["Graph"]["Nodes"]:
            taskList.append(
                Task(
                    name=task["Name"],
                    displayName=task["Name"],
                    taskType=task["Type"],
                    downstreamTasks=self.get_downstream_tasks(
                        task["UniqueId"], tasks["Graph"]
                    ),
                )
            )
        return taskList

    def ingest_pipelines(self) -> Iterable[OMetaDatabaseAndTable]:
        try:
            for workflow in self.glue.list_workflows()["Workflows"]:
                jobs = self.glue.get_workflow(Name=workflow, IncludeGraph=True)[
                    "Workflow"
                ]
                tasks = self.get_tasks(jobs)
                pipeline_ev = Pipeline(
                    id=uuid.uuid4(),
                    name=jobs["Name"],
                    displayName=jobs["Name"],
                    description="",
                    tasks=tasks,
                    service=EntityReference(
                        id=self.pipeline_service.id, type="pipelineService"
                    ),
                )
                yield pipeline_ev
        except Exception as err:
            logger.error(traceback.format_exc())
            logger.error(traceback.print_exc())
            logger.error(err)

    def close(self):
        pass

    def get_status(self) -> SourceStatus:
        return self.status



# generated by datamodel-codegen:
#   filename:  schema/type/profile.json
#   timestamp: 2022-01-16T19:33:03+00:00

from __future__ import annotations

from typing import Optional

from pydantic import AnyUrl, BaseModel


class ImageList(BaseModel):
    image: Optional[AnyUrl] = None
    image24: Optional[AnyUrl] = None
    image32: Optional[AnyUrl] = None
    image48: Optional[AnyUrl] = None
    image72: Optional[AnyUrl] = None
    image192: Optional[AnyUrl] = None
    image512: Optional[AnyUrl] = None


class Profile(BaseModel):
    images: Optional[ImageList] = None



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from enum import Enum
from typing import List, Optional

import requests
from requests.exceptions import HTTPError

from metadata.config.common import ConfigModel
from metadata.ingestion.ometa.credentials import URL, get_api_version

logger = logging.getLogger(__name__)


class RetryException(Exception):
    pass


class APIError(Exception):
    """
    Represent API related error.
    error.status_code will have http status code.
    """

    def __init__(self, error, http_error=None):
        super().__init__(error["message"])
        self._error = error
        self._http_error = http_error

    @property
    def code(self):
        return self._error["code"]

    @property
    def status_code(self):
        http_error = self._http_error
        if http_error is not None and hasattr(http_error, "response"):
            return http_error.response.status_code

    @property
    def request(self):
        if self._http_error is not None:
            return self._http_error.request

    @property
    def response(self):
        if self._http_error is not None:
            return self._http_error.response


class TimeFrame(Enum):
    Day = "1Day"
    Hour = "1Hour"
    Minute = "1Min"
    Sec = "1Sec"


class ClientConfig(ConfigModel):
    """
    :param raw_data: should we return api response raw or wrap it with
                         Entity objects.
    """

    base_url: str
    api_version: Optional[str] = "v1"
    retry: Optional[int] = 3
    retry_wait: Optional[int] = 30
    retry_codes: List[int] = [429, 504]
    auth_token: Optional[str] = None
    auth_header: Optional[str] = None
    raw_data: Optional[bool] = False
    allow_redirects: Optional[bool] = False


class REST(object):
    def __init__(self, config: ClientConfig):
        self.config = config
        self._base_url: URL = URL(self.config.base_url)
        self._api_version = get_api_version(self.config.api_version)
        self._session = requests.Session()
        self._use_raw_data = self.config.raw_data
        self._retry = self.config.retry
        self._retry_wait = self.config.retry_wait
        self._retry_codes = self.config.retry_codes
        self._auth_token = self.config.auth_token

    def _request(
        self, method, path, data=None, base_url: URL = None, api_version: str = None
    ):
        base_url = base_url or self._base_url
        version = api_version if api_version else self._api_version
        url: URL = URL(base_url + "/" + version + path)
        headers = {"Content-type": "application/json"}
        if self._auth_token is not None and self._auth_token != "no_token":
            headers[self.config.auth_header] = self._auth_token
        opts = {
            "headers": headers,
            # Since we allow users to set endpoint URL via env var,
            # human error to put non-SSL endpoint could exploit
            # uncanny issues in non-GET request redirecting http->https.
            # It's better to fail early if the URL isn't right.
            "allow_redirects": self.config.allow_redirects,
        }
        if method.upper() == "GET":
            opts["params"] = data
        else:
            opts["data"] = data

        total_retries = self._retry if self._retry > 0 else 0
        retry = total_retries
        while retry >= 0:
            try:
                logger.debug("URL {}, method {}".format(url, method))
                logger.debug("Data {}".format(opts))
                return self._one_request(method, url, opts, retry)
            except RetryException:
                retry_wait = self._retry_wait * (total_retries - retry + 1)
                logger.warning(
                    "sleep {} seconds and retrying {} "
                    "{} more time(s)...".format(retry_wait, url, retry)
                )
                time.sleep(retry_wait)
                retry -= 1

    def _one_request(self, method: str, url: URL, opts: dict, retry: int):
        """
        Perform one request, possibly raising RetryException in the case
        the response is 429. Otherwise, if error text contain "code" string,
        then it decodes to json object and returns APIError.
        Returns the body json in the 200 status.
        """
        retry_codes = self._retry_codes
        resp = self._session.request(method, url, **opts)
        try:
            resp.raise_for_status()
        except HTTPError as http_error:
            # retry if we hit Rate Limit
            if resp.status_code in retry_codes and retry > 0:
                raise RetryException()
            if "code" in resp.text:
                error = resp.json()
                if "code" in error:
                    raise APIError(error, http_error)
            else:
                raise
        if resp.text != "":
            return resp.json()
        return None

    def get(self, path, data=None):
        return self._request("GET", path, data)

    def post(self, path, data=None):
        return self._request("POST", path, data)

    def put(self, path, data=None):
        return self._request("PUT", path, data)

    def patch(self, path, data=None):
        return self._request("PATCH", path, data)

    def delete(self, path, data=None):
        return self._request("DELETE", path, data)

    def __enter__(self):
        return self

    def close(self):
        self._session.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



#  Copyright 2021 Collate
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This import verifies that the dependencies are available.

import concurrent.futures
from dataclasses import Field, dataclass, field
from typing import Iterable, List, Optional

import confluent_kafka
from confluent_kafka.admin import AdminClient, ConfigResource
from confluent_kafka.schema_registry.schema_registry_client import (
    Schema,
    SchemaRegistryClient,
)

from metadata.config.common import ConfigModel
from metadata.generated.schema.api.data.createTopic import CreateTopicEntityRequest
from metadata.generated.schema.entity.data.topic import SchemaType, Topic
from metadata.generated.schema.entity.services.messagingService import (
    MessagingServiceType,
)
from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.api.common import IncludeFilterPattern, WorkflowContext, logger
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig
from metadata.utils.helpers import get_messaging_service_or_create


@dataclass
class KafkaSourceStatus(SourceStatus):
    topics_scanned: List[str] = field(default_factory=list)
    filtered: List[str] = field(default_factory=list)

    def topic_scanned(self, topic: str) -> None:
        self.topics_scanned.append(topic)

    def dropped(self, topic: str) -> None:
        self.filtered.append(topic)


class KafkaSourceConfig(ConfigModel):
    bootstrap_servers: str = "localhost:9092"
    schema_registry_url: str = "http://localhost:8081"
    consumer_config: dict = {}
    service_name: str
    service_type: str = "Kafka"
    filter_pattern: IncludeFilterPattern = IncludeFilterPattern.allow_all()


@dataclass
class KafkaSource(Source[CreateTopicEntityRequest]):
    config: KafkaSourceConfig
    admin_client: AdminClient
    report: KafkaSourceStatus

    def __init__(
        self,
        config: KafkaSourceConfig,
        metadata_config: MetadataServerConfig,
        ctx: WorkflowContext,
    ):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.status = KafkaSourceStatus()
        self.service = get_messaging_service_or_create(
            config.service_name,
            MessagingServiceType.Kafka.name,
            config.schema_registry_url,
            config.bootstrap_servers.split(","),
            metadata_config,
        )
        self.schema_registry_client = SchemaRegistryClient(
            {"url": self.config.schema_registry_url}
        )
        self.admin_client = AdminClient(
            {
                "bootstrap.servers": self.config.bootstrap_servers,
                "session.timeout.ms": 6000,
            }
        )

    @classmethod
    def create(cls, config_dict, metadata_config_dict, ctx):
        config = KafkaSourceConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(config, metadata_config, ctx)

    def prepare(self):
        pass

    def next_record(self) -> Iterable[CreateTopicEntityRequest]:
        topics = self.admin_client.list_topics().topics
        for t in topics:
            if self.config.filter_pattern.included(t):
                logger.info("Fetching topic schema {}".format(t))
                topic_schema = self._parse_topic_metadata(t)
                topic = CreateTopicEntityRequest(
                    name=t,
                    service=EntityReference(
                        id=self.service.id, type="messagingService"
                    ),
                    partitions=1,
                )
                if topic_schema is not None:
                    topic.schemaText = topic_schema.schema_str
                    if topic_schema.schema_type == "AVRO":
                        topic.schemaType = SchemaType.Avro.name
                    elif topic_schema.schema_type == "PROTOBUF":
                        topic.schemaType = SchemaType.Protobuf.name
                    elif topic_schema.schema_type == "JSON":
                        topic.schemaType = SchemaType.JSON.name
                    else:
                        topic.schemaType = SchemaType.Other.name

                self.status.topic_scanned(topic.name.__root__)
                yield topic
            else:
                self.status.dropped(t)

    def _parse_topic_metadata(self, topic: str) -> Optional[Schema]:
        logger.debug(f"topic = {topic}")
        schema: Optional[Schema] = None
        try:
            registered_schema = self.schema_registry_client.get_latest_version(
                topic + "-value"
            )
            schema = registered_schema.schema
        except Exception as e:
            self.status.warning(topic, f"failed to get schema: {e} for topic {topic}")

        return schema

    def get_status(self):
        return self.status

    def close(self):
        pass
